#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
"""Packet processor."""

import asyncio
from datetime import datetime as dt, timedelta
import logging
from string import printable
from threading import Lock
from types import SimpleNamespace
from typing import ByteString, Optional, Tuple

from serial import SerialException, serial_for_url  # noqa
from serial_asyncio import SerialTransport

from .command import Command, _pkt_header
from .const import (
    DTM_LONG_REGEX,
    MESSAGE_REGEX,
    NON_DEVICE,
    NUL_DEVICE,
    __dev_mode__,
    id_to_address,
)
from .logger import dt_str

BAUDRATE = 115200
READ_TIMEOUT = 0.5
XONXOFF = True

SERIAL_CONFIG = {"baudrate": BAUDRATE, "xonxoff": XONXOFF}

Pause = SimpleNamespace(
    NONE=timedelta(seconds=0),
    MINIMUM=timedelta(seconds=0.01),
    SHORT=timedelta(seconds=0.05),
    DEFAULT=timedelta(seconds=0.15),
    LONG=timedelta(seconds=0.5),
)

# tx (from sent to gwy, to get back from gwy) seems to takes 0.025
DISABLE_QOS_CODE = False
MAX_BUFFER_LEN = 1
MAX_SEND_COUNT = 1
# RETRANS_TIMEOUT = timedelta(seconds=0.03)
# 0.060 gives false +ve for 10E0?
# 0.065 too low when stressed with (e.g.) schedules, log entries
EXPIRY_TIMEOUT = timedelta(seconds=2.0)  # say 0.5


QOS_RETRY_LIMIT = 2
QOS_TIMEOUT_SECS_RQ = timedelta(seconds=0.2)  # 0.2 too low?
QOS_TIMEOUT_SECS_RP = timedelta(seconds=1.0)

_LOGGER = logging.getLogger(__name__)
if True or __dev_mode__:
    _LOGGER.setLevel(logging.DEBUG)
else:
    _LOGGER.setLevel(logging.WARNING)


def extra(dtm, pkt=None):
    _date, _time = dtm[:26].split("T")
    return {
        "date": _date,
        "time": _time,
        "_packet": str(pkt) + " " if pkt else "",
        "error_text": "",
        "comment": "",
    }


def _logger(log_msg, pkt, dtm_now):
    _LOGGER.warning("%s < %s", pkt, log_msg, extra=extra(dtm_now.isoformat(), pkt))


def split_pkt_line(packet_line: str) -> Tuple[str, str, str]:
    # line format: 'datetime packet < parser-message: * evofw3-errmsg # evofw3-comment'
    def _split(text: str, char: str) -> Tuple[str, str]:
        _list = text.split(char, maxsplit=1)
        return _list[0].strip(), _list[1].strip() if len(_list) == 2 else ""

    packet_tmp, comment = _split(packet_line, "#")
    packet_tmp, error = _split(packet_tmp, "*")
    packet, _ = _split(packet_tmp, "<")
    return packet, f"* {error} " if error else "", f"# {comment} " if comment else ""


class Packet:
    """The packet class."""

    def __init__(self, dtm, pkt, raw_pkt) -> None:
        """Create a packet."""
        self.dtm = dtm
        self.date, self.time = dtm.split("T")  # dtm assumed to be valid

        self._pkt_line = pkt
        self._raw_pkt_line = raw_pkt
        self.packet, self.error_text, self.comment = split_pkt_line(pkt)
        self._packet = self.packet + " " if self.packet else ""  # NOTE: hack 4 logging

        self.addrs = [None] * 3
        self.src_addr = self.dst_addr = None

        self._is_valid = None
        self._is_valid = self.is_valid

    def __str__(self) -> str:
        return self.packet if self.packet else ""

    def __repr__(self):
        return str(self._raw_pkt_line if self._raw_pkt_line else self._pkt_line)

    def __eq__(self, other) -> bool:
        if not hasattr(other, "packet"):
            return NotImplemented
        return self.packet == other.packet

    @property
    def is_valid(self) -> Optional[bool]:
        """Return True if a valid packets, otherwise return False/None & log it."""
        # 'good' packets are not logged here, as they may be for silent discarding

        def validate_addresses() -> Optional[bool]:
            """Return True if the address fields are valid (create any addresses)."""
            for idx, addr in enumerate(
                [self.packet[i : i + 9] for i in range(11, 32, 10)]
            ):
                self.addrs[idx] = id_to_address(addr)

            # This check will invalidate these rare pkts (which are never transmitted)
            # ---  I --- --:------ --:------ --:------ 0001 005 00FFFF02FF
            # ---  I --- --:------ --:------ --:------ 0001 005 00FFFF0200
            if not all(
                (
                    self.addrs[0].id not in (NON_DEVICE.id, NUL_DEVICE.id),
                    (self.addrs[1].id, self.addrs[2].id).count(NON_DEVICE.id) == 1,
                )
            ) and not all(
                (
                    self.addrs[2].id not in (NON_DEVICE.id, NUL_DEVICE.id),
                    self.addrs[0].id == self.addrs[1].id == NON_DEVICE.id,
                )
            ):
                return False

            device_addrs = list(filter(lambda x: x.type != "--", self.addrs))

            self.src_addr = device_addrs[0]
            self.dst_addr = device_addrs[1] if len(device_addrs) > 1 else NON_DEVICE

            if self.src_addr.id == self.dst_addr.id:
                self.src_addr = self.dst_addr
            elif self.src_addr.type == self.dst_addr.type:
                # 064  I --- 01:078710 --:------ 01:144246 1F09 003 FF04B5 (invalid)
                return False

            return len(device_addrs) < 3

        if self._is_valid is not None or not self._pkt_line:
            return self._is_valid

        if self.error_text:  # log all packets with an error
            if self.packet:
                _LOGGER.warning("%s < Bad packet: ", self, extra=self.__dict__)
            else:
                _LOGGER.warning("< Bad packet: ", extra=self.__dict__)
            return False

        if not self.packet and self.comment:  # log null packets only if has a comment
            _LOGGER.warning("", extra=self.__dict__)  # normally a warning
            return False

        # TODO: these packets shouldn't go to the packet log, only STDERR?
        if not MESSAGE_REGEX.match(self.packet):
            err_msg = "invalid packet structure"
        elif not validate_addresses():
            err_msg = "invalid packet addresses"
        elif int(self.packet[46:49]) > 48:  # TODO: is 02/I/22C9 > 24?
            err_msg = "excessive payload length"
        elif int(self.packet[46:49]) * 2 != len(self.packet[50:]):
            err_msg = "mismatched payload length"
        else:  # it is a valid packet
            # TODO: Check that an expected RP arrived for an RQ sent by this library
            return True

        _LOGGER.warning("%s < Bad packet: %s ", self, err_msg, extra=self.__dict__)
        return False

    def is_wanted(self, include: list = None, exclude: list = None) -> bool:
        """Silently drop packets with unwanted (e.g. neighbour's) devices.

        Packets to/from HGI80: are never ignored.
        """

        def is_wanted_pkt() -> bool:
            """Return True is a packet is not to be filtered out."""

            if " 18:" in self.packet:  # NOTE: " 18:", leading space is required
                return True
            if include:
                return any(device in self.packet for device in include)
            if exclude:
                return not any(device in self.packet for device in exclude)
            return True

        if is_wanted_pkt():
            _LOGGER.info("%s ", self.packet, extra=self.__dict__)
            return True
        return False

    @property
    def _header(self) -> Optional[str]:
        """Return the QoS header of this packet."""

        if self.is_valid:
            return _pkt_header(str(self))


async def file_pkts(fp):

    for ts_pkt in fp:
        ts_pkt = ts_pkt.strip()
        if ts_pkt == "":  # ignore blank lines
            continue

        try:
            dtm, pkt = ts_pkt[:26], ts_pkt[27:]
            # assuming a completely valid log file, asserts allows for -O for inc. speed
            assert DTM_LONG_REGEX.match(dtm)
            assert dt.fromisoformat(dtm)

        except (AssertionError, TypeError, ValueError):
            _LOGGER.warning(
                "%s < Packet line has an invalid timestamp (ignoring)",
                ts_pkt,
                extra=extra(dt_str(), ts_pkt),
            )
            continue

        pkt = Packet(dtm, pkt, None)
        if pkt.is_valid:  # and pkt.is_wanted(include=include, exclude=exclude):
            yield pkt

        await asyncio.sleep(0)  # usu. 0, only to enable a Ctrl-C


class SerialProtocol(asyncio.Protocol):
    def __init__(self, queue, pkt_handler) -> None:
        # _LOGGER.debug("SerialProtocol.__init__()")

        self._queue = queue
        self._callback = pkt_handler

        self._transport = None
        self._pause_writing = None
        self._recv_buffer = bytes()

        self._qos_lock = Lock()
        self._qos_cmd = None
        self._qos_rp_hdr = None
        self._qos_rq_hdr = None
        self._qos_timeout = None

    def connection_made(self, transport: SerialTransport) -> None:
        """Called when a connection is made."""

        # _LOGGER.debug("SerialProtocol.connection_made(%s)", transport)
        self._transport = transport
        # self._transport.serial.rts = False

    def data_received(self, data: ByteString):
        """Called when some data is received."""

        # _LOGGER.debug("SerialProtocol.data_received(%s)", data)
        # if b'\n' in data:
        #     self._transport.close()

        def create_pkt(pkt_raw: ByteString) -> Packet:
            dtm_str = dt_str()  # done here & now for most-accurate timestamp
            # _LOGGER.debug("%s < Raw pkt", pkt_raw, extra=extra(dtm_str, pkt_raw))

            try:
                pkt_str = "".join(
                    c
                    for c in pkt_raw.decode("ascii", errors="strict").strip()
                    if c in printable
                )
            except UnicodeDecodeError:
                _LOGGER.warning("%s < Bad pkt", pkt_raw, extra=extra(dtm_str, pkt_raw))
                return Packet(dtm_str, "", pkt_raw)

            # any firmware-level packet hacks, i.e. non-HGI80 devices, should be here

            return Packet(dtm_str, pkt_str, pkt_raw)

        def qos_data_received() -> Packet:
            if self._qos_rq_hdr or self._qos_rp_hdr:
                self._qos_lock.acquire()
                if self._qos_rq_hdr == pkt._header:
                    self._qos_rq_hdr = None
                    self._qos_timeout = dt.now() + QOS_TIMEOUT_SECS_RP
                elif self._qos_rp_hdr == pkt._header:
                    self._qos_rp_hdr = None
                    # self._qos_timeout = None
                self._qos_lock.release()

        self._recv_buffer += data
        if b"\r\n" in self._recv_buffer:
            lines = self._recv_buffer.split(b"\r\n")
            self._recv_buffer = lines[-1]

            for line in lines[:-1]:
                pkt = create_pkt(line)
                if pkt.is_valid:
                    self._callback(pkt)
                    qos_data_received()

    async def _write_data(self, data: bytearray) -> None:
        """Called when some data is to be sent (not a callaback)."""

        # _LOGGER.debug("SerialProtocol.send_data(%s)", data)
        while self._pause_writing or not self._transport:
            await asyncio.sleep(0.005)

        while self._transport.serial.out_waiting:
            await asyncio.sleep(0.005)

            # if cmd.verb == " W" or cmd.code in ("0004", "0404", "0418"):
            #     cmd.dtm_timeout = dtm_now + Pause.DEFAULT
            # elif cmd.verb == "RQ":
            #     cmd.dtm_timeout = dtm_now + Pause.SHORT
            # else:
            #     cmd.dtm_timeout = dtm_now + Pause.DEFAULT

        self._transport.write(data)
        # await asyncio.sleep(0.3)  # HACK: until all the other stuff is working

    async def send_data(self, cmd: Command, qos: dict = None) -> None:
        """Called when some data is to be sent (not a callaback)."""

        # self._qos_lock.acquire()
        # self._qos_rq_hdr = self._qos_rp_hdr = None
        # self._qos_lock.release()

        while self._qos_rq_hdr or self._qos_rp_hdr:
            await asyncio.sleep(0.005)
            continue

        await self._write_data(bytearray(f"{cmd}\r\n".encode("ascii")))

        self._qos_lock.acquire()
        self._qos_rq_hdr = cmd._rq_header  # Could be None
        self._qos_rp_hdr = cmd._rp_header  # Could be None, esp. if RQ hdr is None
        self._qos_lock.release()

        if self._qos_rq_hdr:
            self._qos_cmd = cmd
            self._qos_retrys = cmd.qos.get("retry_limit", QOS_RETRY_LIMIT)
            self._qos_timeout = dt.now() + QOS_TIMEOUT_SECS_RQ

        while self._qos_rq_hdr or self._qos_rp_hdr:
            if self._qos_timeout > dt.now():
                await asyncio.sleep(0.005)
                continue

            if self._qos_retrys == 0:
                # print("TIMED OUT - EXPIRED!")
                self._qos_lock.acquire()
                self._qos_rq_hdr = self._qos_rp_hdr = None
                self._qos_lock.release()
                break
            # print("TIMED OUT - RETRANSMITTING!")

            await self._write_data(bytearray(f"{self._qos_cmd}\r\n".encode("ascii")))
            self._qos_retrys -= 1
            self._qos_timeout = dt.now() + QOS_TIMEOUT_SECS_RQ

    def connection_lost(self, exc: Optional[Exception]) -> None:
        """Called when the connection is lost or closed."""

        # _LOGGER.debug("SerialProtocol.connection_lost(%s)", exc)
        if exc is not None:
            pass
        self._transport.loop.stop()

    def pause_writing(self) -> None:
        """Called when the transport's buffer goes over the high-water mark."""

        # _LOGGER.debug("SerialProtocol.pause_writing()")
        print(self._transport.get_write_buffer_size())

        self._pause_writing = True

    def resume_writing(self) -> None:
        """Called when the transport's buffer drains below the low-water mark."""

        # _LOGGER.debug("SerialProtocol.resume_writing()")
        print(self._transport.get_write_buffer_size())

        self._pause_writing = False



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
"""Evohome serial."""

from datetime import timedelta
from functools import total_ordering
import json
import logging
import struct
from types import SimpleNamespace
from typing import Optional
import zlib

from .const import __dev_mode__, CODES_SANS_DOMAIN_ID, COMMAND_FORMAT, HGI_DEVICE
from .logger import dt_now

# SERIAL_PORT = "serial_port"
# CMD_CODE = "cmd_code"
# CMD_VERB = "cmd_verb"
# PAYLOAD = "payload"

# DEVICE_1 = "device_1"
# DEVICE_2 = "device_2"
# DEVICE_3 = "device_3"

RQ_RETRY_LIMIT = 7
RQ_TIMEOUT = 0.03


Priority = SimpleNamespace(LOW=6, DEFAULT=4, HIGH=2, ASAP=0)
Qos = SimpleNamespace(
    AT_MOST_ONCE=0,  # PUB (no handshake)
    AT_LEAST_ONCE=1,  # PUB, ACK (2-way handshake)
    EXACTLY_ONCE=2,  # PUB, REC, REL (FIN) (3/4-way handshake)
)

_LOGGER = logging.getLogger(__name__)
if False and __dev_mode__:
    _LOGGER.setLevel(logging.DEBUG)
else:
    _LOGGER.setLevel(logging.WARNING)


def _pkt_header(packet, response_header=None) -> Optional[str]:
    """Return the QoS header of a packet."""

    verb = packet[4:6]
    if response_header:
        verb = "RP" if verb == "RQ" else " I"  # RQ/RP, or W/I
    code = packet[41:45]
    addr = packet[21:30] if packet[11:13] == "18" else packet[11:20]
    payload = packet[50:]

    header = "|".join((verb, addr, code))

    if code in ("0005", "000C"):  # zone_idx, type
        return "|".join((header, payload[:4]))

    if code == "0404":  # zone_idx, frag_idx
        return "|".join((header, payload[:2] + payload[10:12]))

    if code == "0418":  # log_idx
        return "|".join((header, payload[4:6]))

    if code in CODES_SANS_DOMAIN_ID:  # have no domain_id
        return header

    return "|".join((header, payload[:2]))  # has a domain_id


@total_ordering
class Command:
    """The command class."""

    def __init__(self, verb, dest_addr, code, payload, **kwargs) -> None:
        """Initialise the class."""
        self.verb = verb
        self.from_addr = kwargs.get("from_addr", HGI_DEVICE.id)
        self.dest_addr = dest_addr if dest_addr is not None else self.from_addr
        self.code = code
        self.payload = payload

        # self.pause = None
        self.qos = kwargs.get("qos", {})

        priority = Priority.HIGH if verb in ("0016", "1FC9") else Priority.DEFAULT
        self._priority = kwargs.get("priority", priority)
        self._priority_dtm = dt_now()  # used for __lt__, etc.

        # qos = Qos.AT_LEAST_ONCE if self.verb in ("RQ", " W") else Qos.AT_MOST_ONCE
        # self.qos = kwargs.get("qos", qos)

        # self.dtm_expires = None  # TODO: these 3 shouldn't be instance attributes?
        # self.dtm_timeout = None
        # self.transmit_count = 0

    def __OUT_repr__(self) -> str:
        result = {"packet": str(self)}
        result.update(
            {
                k: v
                for k, v in self.__dict__.items()
                if k not in ("verb", "from_addr", "dest_addr", "code", "payload")
            }
        )
        return json.dumps(result)

    def __str__(self) -> str:
        return COMMAND_FORMAT.format(
            self.verb,
            self.from_addr,
            self.dest_addr,
            self.code,
            int(len(self.payload) / 2),
            self.payload,
        )

    @property
    def _rq_header(self) -> Optional[str]:
        """Return the QoS header of this (request) packet."""

        return _pkt_header(f"... {self}")

    @property
    def _rp_header(self) -> Optional[str]:
        """Return the QoS header of a response packet (if any)."""

        if self._rq_header:  # will be None if RQ header is None
            return _pkt_header(f"... {self}", response_header=True)

    @staticmethod
    def _is_valid_operand(other) -> bool:
        return hasattr(other, "_priority") and hasattr(other, "_priority_dtm")

    def __eq__(self, other) -> bool:
        if not self._is_valid_operand(other):
            return NotImplemented
        return (self._priority, self._priority_dtm) == (
            other._priority,
            other._priority_dtm,
        )

    def __lt__(self, other) -> bool:
        if not self._is_valid_operand(other):
            return NotImplemented
        return (self._priority, self._priority_dtm) < (
            other._priority,
            other._priority_dtm,
        )


class FaultLog:
    """The fault log (of a system) class."""

    def __init__(self, controller, msg=None, **kwargs) -> None:
        """Initialise the class."""
        self._ctl = controller

        self._gwy = controller._gwy
        self._que = controller._que

        self.id = controller.id

        self._fault_log = []

    def add_fault(self, msg) -> None:
        _LOGGER.error("Sched(%s).add_fragment: xxx", self.id)

    def req_schedule(self) -> int:
        _LOGGER.error("Sched(%s).req_schedule: xxx", self.id)

    def req_fragment(self, restart=False) -> int:
        _LOGGER.error("Sched(%s).req_fragment: xxx", self.id)

        if self._gwy.self.config["disable_sending"]:
            return

    def __repr_(self) -> str:
        return json.dumps(self._fault_log)

    def __str_(self) -> str:
        return json.dumps(self._fault_log, indent=2)

    @property
    def fault_log(self) -> list:
        if self._fault_log is not None:
            return self._fault_log

        return self._fault_log


class Schedule:
    """The schedule (of a zone) class."""

    # TODO: stop responding to fragments sent by others
    # TODO: use a lock to only request one schedule at a time

    def __init__(self, zone, msg=None, **kwargs) -> None:
        """Initialise the class."""
        self._zone = zone

        self._ctl = zone._ctl
        self._gwy = zone._gwy
        self._que = zone._que

        self.idx = zone.idx

        # initialse the fragment array: DRY
        self._init_frag_array(total_frags=0)  # could use msg.payload["frag_total"]

        if msg is not None:
            if msg.payload["frag_index"] != 1:
                raise ValueError("not the first fragment of the message")
            self.add_fragment(msg)

    def _init_frag_array(self, total_frags=0) -> None:
        """Reset the fragment array."""
        self.total_frags = total_frags
        self._frag_array = [None] * total_frags
        self._schedule = None

    def add_fragment(self, msg) -> None:
        _LOGGER.error("Sched(%s).add_fragment: xxx", self.idx)

        if msg.code != "0404" or msg.verb != "RP":
            raise ValueError("incorrect message verb/code")
        if msg.payload["zone_idx"] != self.idx:
            raise ValueError("mismatched zone_idx")

        if self.total_frags == 0:
            self._init_frag_array(msg.payload["frag_total"])

        elif self.total_frags != msg.payload["frag_total"]:
            _LOGGER.warning("total fragments has changed: will re-initialise array")
            self._init_frag_array(msg.payload["frag_total"])

        self._frag_array[msg.payload["frag_index"] - 1] = {
            "_msg_dtm": msg.dtm,
            "frag_index": msg.payload["frag_index"],
            "fragment": msg.payload["fragment"],
        }

        # discard any fragments significantly older that this most recent fragment
        for frag in [f for f in self._frag_array if f is not None]:
            # TODO: use a CONST for 5 minutes
            frag = None if frag["_msg_dtm"] < msg.dtm - timedelta(minutes=5) else frag

        # if not [x for x in self._frag_array if x is None]:  # TODO: can leave out?
        #     _ = self.schedule if self._gwy.self.config["disable_sending"] else None

    def req_schedule(self) -> int:
        _LOGGER.error("Sched(%s).req_schedule: xxx", self.idx)

        # TODO: use a lock to ensure only 1 schedule being requested at a time
        # self.req_fragment(restart=True)

    def req_fragment(self, restart=False) -> int:
        """Request the next fragment, and return that fragment's index number.

        Return 0 when there are no more fragments to get.
        """
        _LOGGER.error("Sched(%s).req_fragment: xxx", self.idx)

        if self._gwy.self.config["disable_sending"]:
            return

        if restart is True:  # or self.total_frags == 0
            self._init_frag_array(0)

        if self._frag_array == []:  # aka self.total_frags == 0:
            missing_frags = [0]  # all(frags missing), but how many is unknown
            # kwargs = {"pause": Pause.LONG}  # , "priority": Priority.DEFAULT}

        else:  # aka not all(self._frag_array)
            missing_frags = [i for i, val in enumerate(self._frag_array) if val is None]
            if missing_frags == []:
                print(self.schedule)
                return 0  # not any(frags missing), nothing to add

        header = f"{self.idx}20000800{missing_frags[0] + 1:02d}{self.total_frags:02d}"
        self._que.put_nowait(
            Command("RQ", self._ctl.id, "0404", header, priority=Priority.HIGH)
        )

        return missing_frags[0] + 1

    def __repr_(self) -> str:
        return json.dumps(self._schedule)

    def __str_(self) -> str:
        return json.dumps(self._schedule, indent=2)

    @property
    def schedule(self) -> list:
        if self._schedule is not None:
            return self._schedule

        _LOGGER.warning(
            "Sched(%s).schedule: array is: %s",
            self.idx,
            [{d["frag_index"]: d["fragment"]} for d in self._frag_array],
        )

        if self._frag_array == [] or not all(self._frag_array):
            return

        frags = [v for d in self._frag_array for k, v in d.items() if k == "fragment"]
        try:
            raw_schedule = zlib.decompress(bytearray.fromhex("".join(frags)))
        except zlib.error:
            _LOGGER.exception("Invalid schedule fragments: %s", self._frag_array)
            self._init_frag_array(total_frags=0)
            return

        self._schedule = []
        old_day, switchpoints = 0, []

        for i in range(0, len(raw_schedule), 20):
            zone, day, time, temp, _ = struct.unpack(
                "<xxxxBxxxBxxxHxxHH", raw_schedule[i : i + 20]
            )
            if day > old_day:
                self._schedule.append(
                    {"day_of_week": old_day, "switchpoints": switchpoints}
                )
                old_day, switchpoints = day, []
            switchpoints.append(
                {
                    "time_of_day": "{0:02d}:{1:02d}".format(*divmod(time, 60)),
                    "heat_setpoint": temp / 100,
                }
            )

        self._schedule.append({"day_of_week": old_day, "switchpoints": switchpoints})

        _LOGGER.debug("Sched(%s): len(schedule): %s", self.idx, len(self._schedule))
        # _LOGGER.debug("Sched(%s).schedule: %s", self.idx, self._schedule)
        return self._schedule



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
"""Evohome serial."""

import asyncio
from collections import deque
import json
import logging
import os
from queue import PriorityQueue, Empty
import signal
import sys
from threading import Lock
from typing import Dict, List

from serial import serial_for_url
from serial_asyncio import SerialTransport

from .command import Command
from .const import __dev_mode__, ATTR_ORPHANS
from .devices import DEVICE_CLASSES, Device
from .discovery import start_tests
from .logger import (
    set_logging,
    BANDW_SUFFIX,
    COLOR_SUFFIX,
    CONSOLE_FMT,
    PKT_LOG_FMT,
)
from .message import _LOGGER as msg_logger, Message
from .packet import (
    _LOGGER as pkt_logger,
    Packet,
    file_pkts,
    SerialProtocol,
    SERIAL_CONFIG,
)
from .schema import CONFIG_SCHEMA, KNOWNS_SCHEMA, load_schema

# from .ser2net import Ser2NetServer
from .system import EvoSystem

# TODO: duplicated in schema.py
DONT_CREATE_MESSAGES = 3
DONT_CREATE_ENTITIES = 2
DONT_UPDATE_ENTITIES = 1

_LOGGER = logging.getLogger(__name__)
if False and __dev_mode__:
    _LOGGER.setLevel(logging.DEBUG)
else:
    _LOGGER.setLevel(logging.WARNING)


async def schedule_task(delay, func, *args, **kwargs):
    """Start a coro after delay seconds."""

    async def scheduled_func(delay, func, *args, **kwargs):
        await asyncio.sleep(delay)
        await func(*args, **kwargs)

    asyncio.create_task(scheduled_func(delay, func, *args, **kwargs))


class GracefulExit(SystemExit):
    code = 1


class Gateway:
    """The gateway class."""

    def __init__(self, serial_port, loop=None, **config) -> None:
        """Initialise the class."""

        if config.get("debug_mode"):
            _LOGGER.setLevel(logging.DEBUG)  # should be INFO?
            _LOGGER.debug("Starting evohome_rf, **config = %s", config)

        self.serial_port = serial_port
        self._loop = loop if loop else asyncio.get_running_loop()
        self.config = CONFIG_SCHEMA(config)

        self._protocol = None

        if self.serial_port and self.config.get("input_file"):
            _LOGGER.warning(
                "Serial port specified (%s), so ignoring input file (%s)",
                self.serial_port,
                self.config["input_file"],
            )
            # self.config["input_file"] = None
        elif self.config.get("input_file") is not None:
            self.config["disable_sending"] = True

        # self._execute_cmd = self.config.get("execute_cmd")

        if self.config["reduce_processing"] >= DONT_CREATE_MESSAGES:
            _stream = (None, sys.stdout)
        else:
            _stream = (sys.stdout, None)
        set_logging(msg_logger, stream=_stream[0], file_name=None)
        set_logging(
            pkt_logger,
            stream=_stream[1],
            file_name=self.config.get("packet_log"),
            file_fmt=PKT_LOG_FMT + BANDW_SUFFIX,
            cons_fmt=CONSOLE_FMT + COLOR_SUFFIX,
        )

        self.cmd_que = PriorityQueue()  # TODO: maxsize=200)
        self._buffer = deque()
        self._sched_zone = None
        self._sched_lock = Lock()

        self._prev_msg = None

        self._tasks = []
        self._setup_signal_handler()

        # if config.get("ser2net_server"):
        self._relay = None  # ser2net_server relay

        # if self.config["reduce_processing"] > 0:
        self.evo = None  # EvoSystem(controller=config["controller_id"])
        self.systems: List[EvoSystem] = []
        self.system_by_id: Dict = {}
        self.devices: List[Device] = []
        self.device_by_id: Dict = {}

        self._schema = config.pop("schema", {})
        self.known_devices = {}  # self._include_list + self._exclude_list
        self._known_devices = (
            load_schema(self, self._schema) if self.config["use_schema"] else {}
        )
        self.config["known_devices"] = False  # bool(self.known_devices)

        self._include_list = {}
        self._exclude_list = {}
        if self.config["enforce_allowlist"]:
            self._include_list = KNOWNS_SCHEMA(config.pop("allowlist", {}))
        elif self.config["enforce_blocklist"]:
            self._exclude_list = KNOWNS_SCHEMA(config.pop("blocklist", {}))

        if self.config.get("execute_cmd"):  # e.g. "RQ 01:145038 1F09 00"
            cmd = self.config["execute_cmd"]
            qos = {"retry_limit": 9}
            self.cmd_que.put_nowait(
                Command(cmd[:2], cmd[3:12], cmd[13:17], cmd[18:], qos=qos)
            )

        if True:  # self.config.get("do_discovery"):
            start_tests(self.cmd_que)

    def __repr__(self) -> str:
        return json.dumps(self.schema)

    def __str__(self) -> str:
        return json.dumps(self.schema, indent=2)

    def _setup_signal_handler(self):
        def _sig_handler_win32(signalnum, frame):
            """2 = signal.SIGINT (Ctrl-C)."""
            _LOGGER.info("Received a signal (signalnum=%s), processing...", signalnum)

            if signalnum == signal.SIGINT:  # is this the only useful win32 signal?
                self.cleanup("_sig_handler_win32()")

                raise GracefulExit()

        async def _sig_handler_posix(signal):
            """Handle signals on posix platform."""
            _LOGGER.info("Received a signal (%s), processing...", signal.name)

            if signal in (signal.SIGHUP, signal.SIGINT, signal.SIGTERM):
                self.cleanup("_sig_handler_posix()")  # OK for after tasks.cancel

                tasks = [
                    t for t in asyncio.all_tasks() if t is not asyncio.current_task()
                ]
                [task.cancel() for task in tasks]
                logging.debug(f"Cancelling {len(tasks)} outstanding tasks...")

                # raise CancelledError
                await asyncio.gather(*tasks, return_exceptions=True)

            elif signal == signal.SIGUSR1:
                _LOGGER.info("Params: \r\n%s", {self.evo.id: self.evo.params})

            elif signal == signal.SIGUSR2:
                _LOGGER.info("Status: \r\n%s", {self.evo.id: self.evo.status})

        _LOGGER.debug("Creating signal handlers...")
        signals = [signal.SIGINT, signal.SIGTERM]

        if os.name == "nt":  # TODO: or is sys.platform better?
            for sig in signals + [signal.SIGBREAK]:
                signal.signal(sig, _sig_handler_win32)

        else:  # if os.name == "posix":
            for sig in signals + [signal.SIGHUP, signal.SIGUSR1, signal.SIGUSR2]:
                self._loop.add_signal_handler(
                    sig, lambda sig=sig: asyncio.create_task(_sig_handler_posix(sig))
                )

    def cleanup(self, xxx=None) -> None:
        """Perform the non-async portion of a graceful shutdown."""

        _LOGGER.debug("cleanup() invoked by: %s", xxx)

        if self.config["known_devices"]:
            _LOGGER.debug("cleanup(): Updating known_devices file...")
            for d in self.devices:
                device_attrs = {
                    "friendly_name": d._friendly_name,
                    "ignore": d._ignored,
                }
                if d.id in self.known_devices:
                    self.known_devices[d.id].update(device_attrs)
                else:
                    self.known_devices[d.id] = device_attrs

            with open(self.config["known_devices"], "w") as json_file:
                json.dump(self.known_devices, json_file, sort_keys=True, indent=4)

    @asyncio.coroutine
    def _create_serial_interface(self):
        ser = serial_for_url(self.serial_port, **SERIAL_CONFIG)
        self._protocol = SerialProtocol(self.cmd_que, self._process_packet)
        transport = SerialTransport(self._loop, self._protocol, ser)
        return (transport, self._protocol)

    async def start(self) -> None:
        async def file_reader(fp):
            async for raw_pkt in file_pkts(fp):
                # include=self._include_list, exclude=self._exclude_list
                self._process_packet(raw_pkt)
                await asyncio.sleep(0)  # needed for Ctrl_C to work?

        async def port_writer():
            while True:
                if self.cmd_que.empty():
                    await asyncio.sleep(0.05)
                    continue

                try:
                    cmd = self.cmd_que.get(False)
                except Empty:
                    continue

                if self._protocol:  # or not self.config["disable_sending"]
                    await self._protocol.send_data(cmd)  # put_pkt(cmd, _LOGGER)

                self.cmd_que.task_done()

        # The source of packets is either a serial port, or a text stream
        if self.serial_port:
            reader = asyncio.create_task(self._create_serial_interface())
            writer = asyncio.create_task(port_writer())

            self._tasks.extend([reader, writer])
            await writer

        else:  # if self.config["input_file"]:
            reader = asyncio.create_task(file_reader(self.config["input_file"]))
            writer = asyncio.create_task(port_writer())  # to consume cmds

            self._tasks.extend([reader, writer])
            await reader

        self.cleanup("start()")  # await asyncio.gather(*self._tasks)

    def _process_packet(self, pkt: Packet) -> None:
        """Decode the packet and its payload."""

        if not pkt.is_wanted(include=self._include_list, exclude=self._exclude_list):
            return

        try:
            if self.config["reduce_processing"] >= DONT_CREATE_MESSAGES:
                return

            msg = Message(self, pkt)  # trap/logs all invalids msgs appropriately

            # 18:/RQs are unreliable, the corresponding RPs, if any, are required
            if msg.src.type == "18":
                return

            if self.config["reduce_processing"] >= DONT_CREATE_ENTITIES:
                return

            msg.create_devices()  # from pkt header & from msg payload (e.g. 000C)
            msg.create_zones()  # create zones & ufh_zones (TBD)

            if self.config["reduce_processing"] >= DONT_UPDATE_ENTITIES:
                return

            msg.update_entities(self._prev_msg)  # update the state database

        except (AssertionError, NotImplementedError):
            return

        self._prev_msg = msg if msg.is_valid else None

    def get_device(self, dev_addr, controller=None, domain_id=None) -> Device:
        """Return a device (will create it if required).

        Can also set a controller/system (will create as required). If a controller is
        provided, can also set the domain_id as one of: zone_idx, FF (controllers), FC
        (heater_relay), HW (DHW sensor, relay), or None (unknown, TBA).
        """

        ctl = (
            None if controller is None else self.get_device(controller, domain_id="FF")
        )
        if self.evo is None:
            self.evo = ctl

        if dev_addr.type in ("18", "63", "--"):  # valid addresses, but not devices
            return

        if isinstance(dev_addr, Device):
            device = dev_addr
        else:
            device = self.device_by_id.get(dev_addr.id)

        if device is None:
            if dev_addr.type in ("01", "23") or domain_id == "FF":
                device = EvoSystem(self, dev_addr, domain_id=domain_id)
            else:
                device = DEVICE_CLASSES.get(dev_addr.type, Device)(
                    self, dev_addr, controller=ctl, domain_id=domain_id
                )
        else:  # update the existing device with any metadata
            if ctl is not None:
                device.controller = ctl

            if domain_id in ("FC", "FF"):
                device._domain_id = domain_id
            elif domain_id is not None and ctl is not None:
                device.zone = ctl.get_zone(domain_id)

        return device

    @property
    def schema(self) -> dict:
        """Return the global schema."""

        schema = {"main_controller": self.evo.id if self.evo else None}

        if self.evo:
            schema[self.evo.id] = self.evo.schema
        for evo in self.systems:
            schema[evo.id] = evo.schema

        orphans = [d.id for d in self.devices if d.controller is None]
        orphans.sort()
        schema[ATTR_ORPHANS] = orphans

        return schema

    @property
    def params(self) -> dict:
        result = {}

        result["devices"] = {
            d.id: d.params for d in sorted(self.devices, key=lambda x: x.id)
        }

        return result

    @property
    def status(self) -> dict:
        result = {}

        result["devices"] = {
            d.id: d.status for d in sorted(self.devices, key=lambda x: x.id)
        }

        return result



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
"""Evohome serial."""

from datetime import datetime as dt, timedelta
import logging
from typing import Optional, Union

from .const import (
    ATTR_DHW_VALVE_HTG,
    ATTR_DHW_VALVE,
    CODE_SCHEMA,
    CODE_0005_ZONE_TYPE,
    CODE_000C_DEVICE_TYPE,
    CODE_0418_DEVICE_CLASS,
    CODE_0418_FAULT_STATE,
    CODE_0418_FAULT_TYPE,
    CODES_SANS_DOMAIN_ID,
    DOMAIN_TYPE_MAP,
    MAY_USE_DOMAIN_ID,
    MAY_USE_ZONE_IDX,
    MAX_ZONES,
    SYSTEM_MODE_MAP,
    ZONE_MODE_MAP,
    __dev_mode__,
)
from .devices import dev_hex_to_id
from .opentherm import OPENTHERM_MESSAGES, OPENTHERM_MSG_TYPE, ot_msg_value, parity

_LOGGER = logging.getLogger(__name__)
if __dev_mode__:
    _LOGGER.setLevel(logging.DEBUG)
else:
    _LOGGER.setLevel(logging.WARNING)


def _idx(seqx, msg) -> dict:
    """Determine if a payload has an entity id, usually a domain id or a zone idx.

    Will return either: {"id_name": seqx} or {}.

    The challenge is that payloads starting with (e.g.):
    - "00" are *often not* a zone idx, and
    - "01" *may not* be a zone idx

    Anything in the range F0-FF appears to be a domain id (no false +ve/-ves).
    """
    if msg.code in CODES_SANS_DOMAIN_ID:  # don't idx, even though some != "00"
        # 1F09: "FF" (I), "00" (RP), "F8" (W, after 1FC9)
        # 1FC9: dict is currently encoded in a way that id/idx is not used
        # 2E04: payload[:2] is system mode, would fail final assert
        return {}

    # TODO: these are not evohome, and list of msg codes ?not complete (e.g. 3150?)
    # if {"03", "12", "22"} & {msg.src.type}:
    #     assert seqx == "00"
    #     return {}

    # 045  I --- 03:183434 --:------ 03:183434 1060 003 00FF00
    if {"03", "12", "22"} & {msg.src.type} and msg.src.type == msg.devs[2].type:
        #  msg.code in ("0008", "0009", "1030", "1060", "1100", "2309", "1030", "313F"):
        if msg.code not in ("1030", "2309"):
            assert seqx == "00"
            return {}
        assert int(seqx, 16) < MAX_ZONES
        return {"other_idx": seqx}  # TODO: Should be parent_idx, but still a WIP

    elif msg.code in ("0002", "2D49"):  # non-evohome: hometronics
        return {"other_idx": seqx}

    # TODO: 000C to a UFC should be ufh_ifx, not zone_idx
    elif msg.code == "000C":  # an exception to the usual rules
        if msg.verb == " I":
            return {}
        if msg.raw_payload[2:4] in ("0D", "0E"):  # ("000D", "000E", "010E")
            return {"domain_id": "FA"}
        if msg.raw_payload[2:4] == "0F":
            return {"domain_id": "FC"}
        if msg.src.type == "02":  # in (msg.src.type, msg.dst.type):  # TODO: above
            assert int(seqx, 16) < 8
            if msg.raw_payload[4:6] == "7F":
                return {"Ufh_idx": seqx, "zone_id": None}
            assert int(msg.raw_payload[4:6], 16) < MAX_ZONES
            return {"uFh_idx": seqx, "zone_id": msg.raw_payload[4:6]}
        if msg.dst.type == "02":
            assert int(seqx, 16) < 8
            return {"ufH_idx": seqx}

        assert int(seqx, 16) < MAX_ZONES
        return {"zone_idx": seqx}

    elif msg.code == "0016":  # WIP, not normally {"uses_zone_idx": True}
        if {"12", "22"} & {msg.src.type, msg.dst.type}:
            assert int(seqx, 16) < MAX_ZONES
            idx_name = (
                "zone_idx" if msg.src.type in ("01", "02", "18") else "parent_idx"
            )
            return {idx_name: seqx}

    elif msg.code == "0418":  # does have domain_id/zone_idx, but uses log_idx
        assert int(seqx, 16) < 64  # a 'null' RP has no log_idx == 0
        return {}  # a 'null' RP has no log_idx

    elif msg.code == "22C9":  # these are UFH-specific
        assert int(seqx, 16) < 8  # this can be a "00", maybe zone_idx, see below
        return {"ufh_idx": seqx}  # TODO: confirm is / is not zone_idx

    elif msg.code in ("31D9", "31DA"):  # ventilation
        assert seqx in ("00", "01", "21")
        return {"vent_id": seqx}

    elif msg.code in MAY_USE_DOMAIN_ID and seqx in DOMAIN_TYPE_MAP:
        # no false +ve/-ves, although FF is not a true domain
        return {"domain_id": seqx}

    elif msg.code in MAY_USE_ZONE_IDX:
        assert int(seqx, 16) < MAX_ZONES
        if {"01", "02", "23"} & {msg.src.type, msg.dst.type}:  # to/from a controller
            # if msg.src.type in ("01", "02", "23", "18"):  # This is the old way...
            #     idx_name = "zone_idx"
            # else:
            #     idx_name = "parent_idx"

            # This is the new way...
            if msg.src.type == "02" and msg.src == msg.dst:
                idx_name = "ufh_idx"
            elif msg.src.type in ("01", "02", "23", "18"):
                idx_name = "zone_idx"
            else:
                idx_name = "parent_idx"
            return {idx_name: seqx}

        # 055  I 028 03:094242 --:------ 03:094242 30C9 003 010B22
        elif msg.src.type == "03":  # TODO: WIP
            return {"parent_idx": seqx}  # not zone_idx

    elif msg.code in ("????"):
        assert seqx == "FF"  # only a few "FF"
        return {}

    assert seqx == "00"
    return {}


def parser_decorator(func):
    """Decode the payload (or meta-data) of any message with useful information.

    Also includes some basic payload validation via ASSERTs (e.g payload length).
    """

    def wrapper(*args, **kwargs) -> Optional[dict]:
        """Determine which packets shouldn't be sent through their parser."""

        payload = args[0]
        msg = args[1]

        if msg.verb == " W":  # TODO: WIP, need to check _idx()
            # these are OK to parse Ws:
            if msg.code in ("0001"):
                return {**_idx(payload[:2], msg), **func(*args, **kwargs)}
            # 045  W --- 12:010740 01:145038 --:------ 2309 003 0401F4
            if msg.code in ("2309", "2349") and msg.src.type in ("12", "22", "34"):
                assert int(payload[:2], 16) < MAX_ZONES
                return func(*args, **kwargs)
            # TODO: these are WIP
            if msg.code == "1F09":
                assert payload[:2] == "F8"
                return func(*args, **kwargs)
            if msg.code in ("1FC9"):
                return func(*args, **kwargs)
            # assert payload[:2] in ("00", "FC")  # ("1100", "2309", "2349")
            return func(*args, **kwargs)

        if msg.verb != "RQ":  # i.e. in (" I", "RP")
            result = func(*args, **kwargs)
            if isinstance(result, list):
                return result
            return {**_idx(payload[:2], msg), **result}

        # except for 18:, these should return nothing - 000A is rq_len 1 or 3?
        # grep -E 'RQ.* 002 ' | grep -vE ' (0004|0016|3EF1) '
        # grep -E 'RQ.* 001 ' | grep -vE ' (000A|1F09|22D9|2309|313F|31DA|3EF0) '

        # HACK: to keep logs clean - will need cleaning up eventually
        if msg.src.type == "18" and msg.verb == "RQ":
            if msg.code in ("10A0", "12B0", "2349", "30C9"):
                assert msg.len <= 2
                return {**_idx(payload[:2], msg)}

        # some packets have more than just a domain_id
        if msg.code == "000C":
            assert msg.len == 2
            return {**_idx(payload[:2], msg), **func(*args, **kwargs)}

        if msg.code in ("0004", "000C", "0016", "12B0", "30C9"):
            assert msg.len == 2  # 12B0 will RP to 1
            return {**_idx(payload[:2], msg)}

        if msg.code == "2349":
            assert msg.len == 7
            return {**_idx(payload[:2], msg)}

        if msg.code in ("000A", "2309"):
            if msg.src.type in ("12", "22"):  # is rp_length
                assert msg.len == 6 if msg.code == "000A" else 3
            else:
                assert msg.len == 1  # incl. 34:, rq_length
            return {**_idx(payload[:2], msg)}

        if msg.code == "0005":
            assert msg.len == 2
            return func(*args, **kwargs)  # has no domain_id

        if msg.code == "0100":  # 04: will RQ language
            assert msg.len in (1, 5)  # len(RQ) = 5, but 00 accepted
            return func(*args, **kwargs)  # no context

        if msg.code == "0404":
            return {**_idx(payload[:2], msg), **func(*args, **kwargs)}

        if msg.code == "0418":
            assert msg.len == 3
            assert payload[:4] == "0000"
            assert int(payload[4:6], 16) <= 63
            return {"log_idx": payload[4:6]}

        if msg.code == "10A0":
            # 045 RQ --- 07:045960 01:145038 --:------ 10A0 006 0013740003E4
            # 037 RQ --- 18:013393 01:145038 --:------ 10A0 001 00
            # 054 RP --- 01:145038 18:013393 --:------ 10A0 006 0013880003E8
            assert msg.len == 6 if msg.src.type == "07" else 1
            return func(*args, **kwargs)

        if msg.code == "1100":
            assert payload[:2] in ("00", "FC")
            if msg.len > 2:  # these RQs have payloads!
                return func(*args, **kwargs)
            return {**_idx(payload[:2], msg)}

        if msg.code in ("1260", "10E0", "1F41", "1FC9", "2E04"):  # TODO: Check these
            # These have only been seen when sent by 18:
            assert payload == "FF" if msg.code == "2E04" else "00"  # so: msg.len == 1
            return {}

        if msg.code in ("0008", "1F09", "22D9", "313F", "3EF0"):
            # 061 RQ --- 04:189082 01:145038 --:------ 1F09 001 00
            # 067 RQ --- 01:187666 10:138822 --:------ 22D9 001 00
            # 045 RQ --- 04:056061 01:145038 --:------ 313F 001 00
            # 045 RQ --- 01:158182 13:209679 --:------ 3EF0 001 00
            # 065 RQ --- 01:078710 10:067219 --:------ 3EF0 001 00
            assert payload == "00"  # implies: msg.len == 1
            return {}

        if msg.code in ("31D9", "31DA"):  # ventilation
            # 047 RQ --- 32:168090 30:082155 --:------ 31DA 001 21
            assert msg.len == 1
            return {**_idx(payload[:2], msg)}

        if msg.code == "3220":  # CTL -> OTB (OpenTherm)
            assert msg.len == 5
            return func(*args, **kwargs)

        if msg.code == "3EF1":
            # 082 RQ --- 31:110943 13:068890 --:------ 3EF1 001 00
            # 082 RQ --- 22:091267 01:140959 --:------ 3EF1 002 0700
            # 088 RQ --- 22:054901 13:133379 --:------ 3EF1 002 0000
            if msg.len > 1:
                assert payload[2:] == "00"  # implies: msg.len >= 2
            return {**_idx(payload[:2], msg)}

        if msg.code in CODE_SCHEMA:
            assert False, "unknown RQ"

        if msg.src.type != "18":
            assert False, "unknown RQ"

    return wrapper


def _bool(value: str) -> Optional[bool]:  # either 00 or C8
    """Return a boolean."""
    assert value in ("00", "C8", "FF")
    return {"00": False, "C8": True}.get(value)


def _dtm(value: str) -> str:
    """Return a local datetime hex string in isoformat."""
    #        00141B0A07E3  (...HH:MM:00)    for system_mode, zone_mode (schedules?)
    #      0400041C0A07E3  (...HH:MM:SS)    for sync_datetime

    if len(value) == 12:
        value = f"00{value}"
    assert len(value) == 14
    return dt(
        year=int(value[10:14], 16),
        month=int(value[8:10], 16),
        day=int(value[6:8], 16),
        hour=int(value[4:6], 16) & 0b11111,  # 1st 3 bits: DayOfWeek
        minute=int(value[2:4], 16),
        second=int(value[:2], 16) & 0b1111111,  # 1st bit: used for DST
    ).strftime("%Y-%m-%d %H:%M:%S")


def _date(value: str) -> Optional[str]:  # YY-MM-DD
    """Return a date string in the format YY-MM-DD."""
    assert len(value) == 8
    if value == "FFFFFFFF":
        return
    return dt(
        year=int(value[4:8], 16),
        month=int(value[2:4], 16),
        day=int(value[:2], 16) & 0b11111,  # 1st 3 bits: DayOfWeek
    ).strftime("%Y-%m-%d")


def _percent(value: str) -> Optional[float]:  # a percentage 0-100% (0.0 to 1.0)
    """Return a percentage, 0-100% with resolution of 0.5%."""
    assert len(value) == 2
    if value in ("FE", "FF"):  # TODO: diff b/w FE (seen with 3150) & FF
        return
    assert int(value, 16) <= 200
    return int(value, 16) / 200


def _str(value: str) -> Optional[str]:  # printable ASCII characters
    """Return a string of printable ASCII characters."""
    _string = bytearray([x for x in bytearray.fromhex(value) if 31 < x < 127])
    return _string.decode("ascii") if _string else None


def _temp(value: str) -> Union[float, bool, None]:
    """Return a two's complement Temperature/Setpoint."""
    assert len(value) == 4
    if value == "31FF":  # means: N/A (== 127.99, 2s complement)
        return
    if value == "7EFF":  # possibly only for setpoints?
        return False
    if value == "7FFF":  # also: FFFF?, means: N/A (== 327.67)
        return
    temp = int(value, 16)
    return (temp if temp < 2 ** 15 else temp - 2 ** 16) / 100


@parser_decorator  # rf_unknown
def parser_0001(payload, msg) -> Optional[dict]:
    # When in test mode, a 12: will send a W every 6 seconds, *on?* the second:
    # 12:39:56.099 061  W --- 12:010740 --:------ 12:010740 0001 005 0000000501
    # 12:40:02.098 061  W --- 12:010740 --:------ 12:010740 0001 005 0000000501
    # 12:40:08.099 058  W --- 12:010740 --:------ 12:010740 0001 005 0000000501

    # sent by a THM every 5s when is signal strength test mode (0505, except 1st pkt)
    # 13:48:38.518 080  W --- 12:010740 --:------ 12:010740 0001 005 0000000501
    # 13:48:45.518 074  W --- 12:010740 --:------ 12:010740 0001 005 0000000505
    # 13:48:50.518 077  W --- 12:010740 --:------ 12:010740 0001 005 0000000505

    # sent by a CTL before a rf_check
    # 15:12:47.769 053  W --- 01:145038 --:------ 01:145038 0001 005 FC00000505
    # 15:12:47.869 053 RQ --- 01:145038 13:237335 --:------ 0016 002 00FF
    # 15:12:47.880 053 RP --- 13:237335 01:145038 --:------ 0016 002 0017

    # 12:30:18.083 047  W --- 01:145038 --:------ 01:145038 0001 005 0800000505
    # 12:30:23.084 049  W --- 01:145038 --:------ 01:145038 0001 005 0800000505

    # 15:03:33.187 054  W --- 01:145038 --:------ 01:145038 0001 005 FC00000505
    # 15:03:38.188 063  W --- 01:145038 --:------ 01:145038 0001 005 FC00000505
    # 15:03:43.188 064  W --- 01:145038 --:------ 01:145038 0001 005 FC00000505
    # 15:13:19.757 053  W --- 01:145038 --:------ 01:145038 0001 005 FF00000505
    # 15:13:24.758 054  W --- 01:145038 --:------ 01:145038 0001 005 FF00000505
    # 15:13:29.758 068  W --- 01:145038 --:------ 01:145038 0001 005 FF00000505
    # 15:13:34.759 063  W --- 01:145038 --:------ 01:145038 0001 005 FF00000505

    # loopback (not Tx'd) by a HGI80 whenever its button is pressed
    # 00:22:41.540 ---  I --- --:------ --:------ --:------ 0001 005 00FFFF02FF
    # 00:22:41.757 ---  I --- --:------ --:------ --:------ 0001 005 00FFFF0200
    # 00:22:43.320 ---  I --- --:------ --:------ --:------ 0001 005 00FFFF02FF
    # 00:22:43.415 ---  I --- --:------ --:------ --:------ 0001 005 00FFFF0200

    assert msg.verb in (" I", " W")
    assert msg.len == 5
    assert payload[:2] in ("FC", "FF") or (int(payload[:2], 16) < MAX_ZONES)
    assert payload[2:6] in ("0000", "FFFF")
    assert payload[6:8] in ("02", "05")
    return {
        **_idx(payload[:2], msg),  # not fully understood
        "unknown_0": payload[2:6],
        "unknown_1": payload[6:8],
        "unknown_2": payload[8:],
    }


@parser_decorator  # sensor_weather
def parser_0002(payload, msg) -> Optional[dict]:
    assert msg.len == 4

    return {
        **_idx(payload[:2], msg),
        "temperature": _temp(payload[2:6]),
        "unknown_0": payload[6:],
    }


@parser_decorator  # zone_name
def parser_0004(payload, msg) -> Optional[dict]:
    # RQ payload is zz00; limited to 12 chars in evohome UI? if "7F"*20: not a zone

    assert msg.len == 22
    assert payload[2:4] == "00"

    if payload[4:] == "7F" * 20:
        return {**_idx(payload[:2], msg)}

    return {**_idx(payload[:2], msg), "name": _str(payload[4:])}


@parser_decorator  # system_zone (add/del a zone?)
def parser_0005(payload, msg) -> Optional[dict]:
    # RQ payload is xx00, controller wont respond to a xx

    # 047  I --- 34:064023 --:------ 34:064023 0005 012 000A0000 000F0000 00100000
    # 045  I --- 01:145038 --:------ 01:145038 0005 004 00000100

    def _parser(seqx) -> dict:
        def _get_flag8(byte, *args) -> list:
            """Split a byte (as a str) into a list of 8 bits (1/0)."""
            ret = [0] * 8
            byte = bytes.fromhex(byte)[0]
            for i in range(0, 8):
                ret[i] = byte & 1
                byte = byte >> 1
            return ret

        # assert seqx[:2] == "00"  # done in _idx
        assert len(seqx) == 8
        assert payload[2:4] in CODE_0005_ZONE_TYPE

        return {
            "zone_mask": (_get_flag8(seqx[4:6]) + _get_flag8(seqx[6:8]))[:MAX_ZONES],
            "zone_type": CODE_0005_ZONE_TYPE.get(seqx[2:4], seqx[2:4]),
        }

    if msg.verb == "RQ":
        assert payload[:2] == "00"
        return {
            "zone_type": CODE_0005_ZONE_TYPE.get(payload[2:4], payload[2:4]),
        }

    assert msg.verb in (" I", "RP")
    if msg.src.type == "34":
        assert msg.len == 12  # or % 4?
        return [_parser(payload[i : i + 8]) for i in range(0, len(payload), 8)]

    assert msg.src.type in ("01", "02")  # and "23"?
    return _parser(payload)


@parser_decorator  # schedule_sync (any changes?)
def parser_0006(payload, msg) -> Optional[dict]:
    """Return number of changes to the schedules (not fully understood).

    Each change increments the counter by 2. Includes DHW schedule.
    """
    # 16:10:34.288 053 RQ --- 18:013393 01:145038 --:------ 0006 001 00
    # 16:10:34.291 053 RP --- 01:145038 18:013393 --:------ 0006 004 0005 0008
    #              --- RQ --- 30:071715 01:067930 --:------ 0006 001 00

    if msg.verb == "RQ":
        assert payload == "00"  # msg.len == 1
        return {}

    assert msg.verb == "RP"
    assert msg.len == 4
    assert payload[:2] == "00"  # otherwise: payload[2:] == "FFFFFF", invalid
    assert payload[2:4] in ("05", "FF")

    return {"header": payload[:4], "num_changes": int(payload[4:], 16)}


@parser_decorator  # relay_demand (domain/zone/device)
def parser_0008(payload, msg) -> Optional[dict]:
    # https://www.domoticaforum.eu/viewtopic.php?f=7&t=5806&start=105#p73681
    # e.g. Electric Heat Zone
    assert msg.len == 2

    if payload[:2] not in ("F9", "FA", "FC"):
        assert int(payload[:2], 16) < MAX_ZONES  # TODO: when 0, when FC, when zone

    return {**_idx(payload[:2], msg), "relay_demand": _percent(payload[2:])}


@parser_decorator  # relay_failsafe
def parser_0009(payload, msg) -> Union[dict, list]:
    # TODO: can only be max one relay per domain/zone
    # can get: 003 or 006, e.g.: FC01FF-F901FF or FC00FF-F900FF
    # 095  I --- 23:100224 --:------ 23:100224 0009 003 0100FF  # 2-zone ST9520C

    def _parser(seqx) -> dict:
        assert seqx[:2] in ("F9", "FC") or int(seqx[:2], 16) < MAX_ZONES
        assert seqx[2:4] in ("00", "01")
        assert seqx[4:] in ("00", "FF")

        return {
            **_idx(seqx[:2], msg),
            "failsafe_enabled": {"00": False, "01": True}.get(seqx[2:4]),
        }

    if msg.is_array:
        assert msg.len >= 3 and msg.len % 3 == 0  # assuming not RQ
        return [_parser(payload[i : i + 6]) for i in range(0, len(payload), 6)]

    assert msg.len == 3
    return _parser(payload)


@parser_decorator  # zone_config (zone/s)
def parser_000a(payload, msg) -> Union[dict, list, None]:
    # 11:21:10.674 063 RQ --- 34:044203 01:158182 --:------ 000A 001 08
    # 11:21:10.736 045 RP --- 01:158182 34:044203 --:------ 000A 006 081001F409C4
    # following for 12: too:
    # 13:13:08.273 045 RQ --- 22:017139 01:140959 --:------ 000A 006 080001F40DAC
    # 13:13:08.288 045 RP --- 01:140959 22:017139 --:------ 000A 006 081001F40DAC

    def _parser(seqx) -> dict:
        # if seqx[2:] == "007FFF7FFF":  # (e.g. RP) a null zone

        bitmap = int(seqx[2:4], 16)
        return {
            **_idx(seqx[:2], msg),
            "min_temp": _temp(seqx[4:8]),
            "max_temp": _temp(seqx[8:]),
            "local_override": not bool(bitmap & 1),
            "openwindow_function": not bool(bitmap & 2),
            "multiroom_mode": not bool(bitmap & 16),
            "_unknown_bitmap": f"0b{bitmap:08b}",
        }  # cannot determine zone_type from this information

    if msg.is_array:  # TODO: these msgs can require 2 pkts!
        assert msg.len >= 6 and msg.len % 6 == 0  # assuming not RQ
        return [_parser(payload[i : i + 12]) for i in range(0, len(payload), 12)]

    assert msg.len == 6
    return _parser(payload)


@parser_decorator  # zone_actuators (not sensors)
def parser_000c(payload, msg) -> Optional[dict]:
    # 045  I --- 34:092243 --:------ 34:092243 000C 018 000A7FFFFFFF000F7FFFFFFF00107FFFFFFF  # noqa: E501
    # 045 RP --- 01:145038 18:013393 --:------ 000C 006 00000010DAFD
    # 045 RP --- 01:145038 18:013393 --:------ 000C 012 01000010DAF501000010DAFB

    # RQ payload is zz00, NOTE: aggregation of parsing taken here
    def _parser(seqx) -> dict:
        assert seqx[:2] == payload[:2]
        assert seqx[2:4] in CODE_000C_DEVICE_TYPE
        assert seqx[4:6] == "7F" or int(seqx[4:6], 16) < MAX_ZONES

        # print({dev_hex_to_id(seqx[6:12]): seqx[4:6]})
        return {dev_hex_to_id(seqx[6:12]): seqx[4:6]}

    if msg.verb == "RQ":
        assert msg.len == 2
    else:
        assert msg.len >= 6 and msg.len % 6 == 0  # assuming not RQ

    device_class = CODE_000C_DEVICE_TYPE[payload[2:4]]
    if device_class == ATTR_DHW_VALVE and msg.raw_payload[:2] == "01":
        device_class = ATTR_DHW_VALVE_HTG

    if msg.verb == "RQ":
        return {**_idx(payload[:2], msg), "device_class": device_class}

    devices = [_parser(payload[i : i + 12]) for i in range(0, len(payload), 12)]

    return {
        # **_idx(payload[:2], msg),
        "device_class": device_class,
        "devices": [k for d in devices for k, v in d.items() if v != "7F"],
    }  # TODO: the assumption that all domain_id/zones_idx are the same is wrong


@parser_decorator  # unknown, from STA
def parser_000e(payload, msg) -> Optional[dict]:
    assert payload == "000014"  # rarely, from STA:xxxxxx
    return {"unknown_0": payload}


@parser_decorator  # rf_check
def parser_0016(payload, msg) -> Optional[dict]:
    # TODO: does 0016 include parent_idx
    # 09:05:33.178 046 RQ --- 22:060293 01:078710 --:------ 0016 002 0200
    # 09:05:33.194 064 RP --- 01:078710 22:060293 --:------ 0016 002 021E
    # 12:47:25.080 048 RQ --- 12:010740 01:145038 --:------ 0016 002 0800
    # 12:47:25.094 045 RP --- 01:145038 12:010740 --:------ 0016 002 081E

    assert msg.verb in ("RQ", "RP")
    assert msg.len == 2  # for both RQ/RP, but RQ/00 will work
    # assert payload[:2] == "00"  # e.g. RQ/22:/0z00 (parent_zone), but RQ/07:/0000?

    if msg.verb == "RQ":
        return {}  # {"rf_request": msg.dst.id}

    rf_value = int(payload[2:4], 16)
    return {
        #  "rf_source": msg.dst.id,
        "rf_strength": min(int(rf_value / 5) + 1, 5),
        "rf_value": rf_value,
    }


@parser_decorator  # language (of device/system)
def parser_0100(payload, msg) -> Optional[dict]:
    if msg.verb == "RQ" and payload == "00":  # HACK: should be "00ssssFFFF"
        return {}

    assert msg.len == 5
    assert payload[:2] == "00"
    assert payload[6:] == "FFFF"
    return {"language": _str(payload[2:6]), "_unknown_0": payload[6:]}


@parser_decorator  # unknown, from a HR91 (when its buttons are pushed)
def parser_01d0(payload, msg) -> Optional[dict]:
    # 23:57:28.869 045  W --- 04:000722 01:158182 --:------ 01D0 002 0003
    # 23:57:28.931 045  I --- 01:158182 04:000722 --:------ 01D0 002 0003
    # 23:57:31.581 048  W --- 04:000722 01:158182 --:------ 01E9 002 0003
    # 23:57:31.643 045  I --- 01:158182 04:000722 --:------ 01E9 002 0000
    # 23:57:31.749 050  W --- 04:000722 01:158182 --:------ 01D0 002 0000
    # 23:57:31.811 045  I --- 01:158182 04:000722 --:------ 01D0 002 0000
    assert msg.len == 2
    assert payload[2:] in ("00", "03")
    return {"unknown_0": payload[2:]}


@parser_decorator  # unknown, from a HR91 (when its buttons are pushed)
def parser_01e9(payload, msg) -> Optional[dict]:
    # 23:57:31.581348 048  W --- 04:000722 01:158182 --:------ 01E9 002 0003
    # 23:57:31.643188 045  I --- 01:158182 04:000722 --:------ 01E9 002 0000
    assert msg.len == 2
    assert payload[2:] in ("00", "03")
    return {"unknown_0": payload[2:]}


@parser_decorator  # zone_schedule (fragment)
def parser_0404(payload, msg) -> Optional[dict]:
    def _header(seqx) -> dict:
        assert seqx[2:8] == "200008"

        return {
            # **_idx(payload[:2], msg),  # added by wrapper
            "frag_index": int(seqx[10:12], 16),
            "frag_total": int(seqx[12:], 16),
            "frag_length": int(seqx[8:10], 16),
        }

    if msg.verb == "RQ":
        assert msg.len == 7
        return _header(payload[:14])

    assert msg.verb == "RP"
    return {**_header(payload[:14]), "fragment": payload[14:]}


@parser_decorator  # system_fault
def parser_0418(payload, msg=None) -> Optional[dict]:
    """10 * 6 log entries in the UI, but 63 via RQs."""

    # 045 RP --- 01:145038 18:013393 --:------ 0418 022 000000B00401010000008694A3CC7FFFFF70000ECC8A  # noqa
    # 045 RP --- 01:145038 18:013393 --:------ 0418 022 00C001B004010100000086949BCB7FFFFF70000ECC8A  # noqa
    # 045 RP --- 01:145038 18:013393 --:------ 0418 022 000000B0000000000000000000007FFFFF7000000000  # noqa

    def _timestamp(seqx):
        """In the controller UI: YYYY-MM-DD HH:MM."""
        _seqx = int(seqx, 16)
        return dt(
            year=(_seqx & 0b1111111 << 24) >> 24,
            month=(_seqx & 0b1111 << 36) >> 36,
            day=(_seqx & 0b11111 << 31) >> 31,
            hour=(_seqx & 0b11111 << 19) >> 19,
            minute=(_seqx & 0b111111 << 13) >> 13,
            second=(_seqx & 0b111111 << 7) >> 7,
        ).strftime("%Y-%m-%dT%H:%M:%S")

    if payload == CODE_SCHEMA["0418"]["null_rp"]:
        # a null log entry, or: is payload[38:] == "000000" sufficient?
        return {}
    #
    if msg:
        assert msg.verb in (" I", "RP")
        assert msg.len == 22
    else:
        assert len(payload) / 2 == 22
    #
    assert payload[:2] == "00"  # likely always 00
    assert payload[2:4] in list(CODE_0418_FAULT_STATE)  # C0 doesn't appear in the UI?
    assert int(payload[4:6], 16) <= 63  # TODO: upper limit is: 60? 63? more?
    assert payload[6:8] == "B0"  # unknown_1, ?priority
    assert payload[8:10] in list(CODE_0418_FAULT_TYPE)

    # domain_id == '1C' (should be 'FC'?) seen with below (from evo ctl UI):
    # "FAULT | 28-08-2020, 03:15 | COMMS FAULT, ACTUATOR"
    # {
    #   'timestamp':    '20-08-28T03:15:24',
    #   'fault_state':  'fault',
    #   'fault_type':   'comms_fault',
    #   'device_class': 'actuator',
    #   'domain_id':    '1C',         # should be FC?
    #   'device_id':    '13:163733'   # acting as boiler-relay
    # }
    assert int(payload[10:12], 16) < MAX_ZONES or (
        payload[10:12] in ("F9", "FA", "FC", "1C")
    )
    assert payload[12:14] in list(CODE_0418_DEVICE_CLASS)
    assert payload[14:18] == "0000"  # unknown_2
    assert payload[28:30] in ("7F", "FF")  # TODO: last bit in dt field, DST?
    assert payload[30:38] == "FFFF7000"  # unknown_3

    result = {  # TODO: stop using __idx()?
        "log_idx": payload[4:6],
        "timestamp": _timestamp(payload[18:30]),
        "fault_state": CODE_0418_FAULT_STATE.get(payload[2:4], payload[2:4]),
        "fault_type": CODE_0418_FAULT_TYPE.get(payload[8:10], payload[8:10]),
        "device_class": CODE_0418_DEVICE_CLASS.get(payload[12:14], payload[12:14]),
    }  # TODO: stop using __idx()?

    if payload[12:14] != "00":  # Controller
        key_name = "zone_id" if int(payload[10:12], 16) < MAX_ZONES else "domain_id"
        result.update({key_name: payload[10:12]})  # TODO: don't use zone_idx (for now)

    if payload[38:] == "000002":  # "00:000002 for Unknown?
        result.update({"device_id": None})
    elif payload[38:] not in ("000000", "000001"):  # "00:000001 for Controller?
        result.update({"device_id": dev_hex_to_id(payload[38:])})

    return result


@parser_decorator  # unknown, from STA
def parser_042f(payload, msg) -> Optional[dict]:
    # 055  I --- 34:064023 --:------ 34:064023 042F 008 00000000230023F5
    # 063  I --- 34:064023 --:------ 34:064023 042F 008 00000000240024F5
    # 049  I --- 34:064023 --:------ 34:064023 042F 008 00000000250025F5
    # 045  I --- 34:064023 --:------ 34:064023 042F 008 00000000260026F5
    # 045  I --- 34:092243 --:------ 34:092243 042F 008 0000010021002201
    # 000  I     34:011469 --:------ 34:011469 042F 008 00000100030004BC

    assert msg.len in (8, 9)  # non-evohome are 9
    assert payload[:2] == "00"

    return {
        "counter_1": int(payload[2:6], 16),
        "counter_2": int(payload[6:10], 16),
        "counter_total": int(payload[10:14], 16),
        "unknown_0": payload[14:],
    }


@parser_decorator  # mixvalve_config (zone)
def parser_1030(payload, msg) -> Optional[dict]:
    def _parser(seqx) -> dict:
        assert seqx[2:4] == "01"

        param_name = {
            "C8": "max_flow_temp",
            "C9": "pump_rum_time",
            "CA": "actuator_run_time",
            "CB": "min_flow_temp",
            "CC": "unknown_0",  # ?boolean?
        }[seqx[:2]]

        return {param_name: int(seqx[4:], 16)}

    assert msg.len == 1 + 5 * 3
    assert payload[30:] in ("00", "01")

    params = [_parser(payload[i : i + 6]) for i in range(2, len(payload), 6)]
    return {**_idx(payload[:2], msg), **{k: v for x in params for k, v in x.items()}}


@parser_decorator  # device_battery (battery_state)
def parser_1060(payload, msg) -> Optional[dict]:
    """Return the battery state.

    Some devices (04:) will also report battery level.
    """
    # 06:48:23.948 049  I --- 12:010740 --:------ 12:010740 1060 003 00FF01
    # 16:18:43.515 051  I --- 12:010740 --:------ 12:010740 1060 003 00FF00
    # 16:14:44.180 054  I --- 04:056057 --:------ 04:056057 1060 003 002800
    # 17:34:35.460 087  I --- 04:189076 --:------ 01:145038 1060 003 026401

    assert msg.len == 3
    assert payload[4:6] in ("00", "01")

    return {"low_battery": payload[4:] == "00", "battery_level": _percent(payload[2:4])}


@parser_decorator  # unknown (non-Evohome, e.g. ST9520C)
def parser_1090(payload, msg) -> dict:
    # 14:08:05.176 095 RP --- 23:100224 22:219457 --:------ 1090 005 007FFF01F4
    # 18:08:05.809 095 RP --- 23:100224 22:219457 --:------ 1090 005 007FFF01F4

    # this is an educated guess
    assert msg.len == 5
    assert int(payload[:2], 16) < 2

    return {
        **_idx(payload[:2], msg),
        "temp_0": _temp(payload[2:6]),
        "temp_1": _temp(payload[6:10]),
    }


@parser_decorator  # dhw_params
def parser_10a0(payload, msg) -> Optional[dict]:
    """Return the DHW parameters.

    DHW sends a RQ (not an I) with payload!
    """
    # 057 RQ --- 07:045960 01:145038 --:------ 10A0 006 00-31FF-00-31FF (null)
    # 053 RQ --- 07:045960 01:145038 --:------ 10A0 006 00-1770-00-03E8
    # 049 RQ --- 07:045960 01:145038 --:------ 10A0 006 00-1374-00-03E4
    # 045 RQ --- 07:030741 01:102458 --:------ 10A0 006 00-181F-00-03E4
    # 095 RQ --- 07:036831 23:100224 --:------ 10A0 006 01-1566-00-03E4 (non-evohome)

    assert msg.len == 6
    # assert payload[:2] == "00"  # TODO: all *evohome* DHW pkts have no domain

    return {
        "setpoint": _temp(payload[2:6]),  # 30.0-85.0 C
        "overrun": int(payload[6:8], 16),  # 0-10 minutes
        "differential": _temp(payload[8:12]),  # 1.0-10.0 C
    }


@parser_decorator  # device_info
def parser_10e0(payload, msg) -> Optional[dict]:
    assert msg.len in (30, 36, 38)  # a non-evohome seen with 30

    return {  # TODO: add version?
        "description": _str(payload[36:]),
        "firmware": _date(payload[20:28]),  # could be 'FFFFFFFF'
        "manufactured": _date(payload[28:36]),
        "unknown": payload[:20],
    }


@parser_decorator  # tpi_params (domain/zone/device)
def parser_1100(payload, msg) -> Optional[dict]:
    assert msg.len in (5, 8)
    assert payload[:2] in ("00", "FC")
    # 2020-09-23T19:25:04.767331 047  I --- 13:079800 --:------ 13:079800 1100 008 00170498007FFF01  # noqa
    assert int(payload[2:4], 16) / 4 in range(1, 13)
    assert int(payload[4:6], 16) / 4 in range(1, 31)
    assert int(payload[6:8], 16) / 4 in range(0, 16)
    assert payload[8:10] in ("00", "FF")

    # for TPI
    #  - cycle_rate: 3, 6, 9, 12??
    #  - min_on_time: 1-5??
    #  - min_off_time: ??
    # for heatpump
    #  - cycle_rate: 1-9
    #  - min_on_time: 1, 5, 10,...30
    #  - min_off_time: 0, 5, 10, 15

    def _parser(seqx) -> dict:
        return {
            **_idx(seqx[:2], msg),
            "cycle_rate": int(payload[2:4], 16) / 4,  # in cycles/hour
            "minimum_on_time": int(payload[4:6], 16) / 4,  # in minutes
            "minimum_off_time": int(payload[6:8], 16) / 4,  # in minutes
            # "_unknown_0": payload[8:10],  # always 00, FF?
        }

    if msg.len == 5:
        return _parser(payload)

    assert payload[14:] == "01"
    return {
        **_parser(payload[:10]),
        "proportional_band_width": _temp(payload[10:14]),  # in degrees C
        # "_unknown_1": payload[14:],  # always 01?
    }


@parser_decorator  # dhw_temp
def parser_1260(payload, msg) -> Optional[dict]:
    assert msg.len == 3
    assert payload[:2] == "00"  # all DHW pkts have no domain

    return {"temperature": _temp(payload[2:])}


@parser_decorator  # outdoor_temp
def parser_1290(payload, msg) -> Optional[dict]:
    # evohome responds to an RQ
    assert msg.len == 3
    assert payload[:2] == "00"  # no domain

    return {"temperature": _temp(payload[2:])}


@parser_decorator  # indoor_humidity (Nuaire RH sensor)
def parser_12a0(payload, msg) -> Optional[dict]:
    assert msg.len == 6
    assert payload[:2] == "00"  # domain?

    return {
        "relative_humidity": int(payload[2:4], 16) / 100,  # is not /200
        "temperature": _temp(payload[4:8]),
        "dewpoint_temp": _temp(payload[8:12]),
    }


@parser_decorator  # window_state (of a device/zone)
def parser_12b0(payload, msg) -> Optional[dict]:
    assert payload[2:] in ("0000", "C800", "FFFF")  # "FFFF" means N/A
    # assert msg.len == 3  # implied

    # TODO: zone.open_window = any(TRV.open_windows)?
    return {**_idx(payload[:2], msg), "window_open": _bool(payload[2:4])}


@parser_decorator  # system_sync
def parser_1f09(payload, msg) -> Optional[dict]:
    # TODO: Try RQ/1F09/"F8-FF" (CTL will RP to a RQ/00)
    assert msg.len == 3
    assert payload[:2] in ("00", "F8", "FF")  # W uses F8, non-Honeywell devices use 00

    seconds = int(payload[2:6], 16) / 10
    next_sync = msg.dtm + timedelta(seconds=seconds)

    return {
        "remaining_seconds": seconds,
        "_next_sync": dt.strftime(next_sync, "%H:%M:%S"),
    }


@parser_decorator  # dhw_mode
def parser_1f41(payload, msg) -> Optional[dict]:
    assert msg.len in (6, 12)
    assert payload[:2] == "00"  # all DHW pkts have no domain

    # 053 RP --- 01:145038 18:013393 --:------ 1F41 006 00FF00FFFFFF  # no stored DHW
    assert payload[2:4] in ("00", "01", "FF")
    assert payload[4:6] in list(ZONE_MODE_MAP)
    if payload[4:6] == "04":
        assert msg.len == 12
        assert payload[6:12] == "FFFFFF"

    return {
        "active": {"00": False, "01": True, "FF": None}[payload[2:4]],
        "dhw_mode": ZONE_MODE_MAP.get(payload[4:6]),
        "until": _dtm(payload[12:24]) if payload[4:6] == "04" else None,
    }


@parser_decorator  # rf_bind
def parser_1fc9(payload, msg) -> Optional[dict]:
    # this is an array of codes
    # 049  I --- 01:145038 --:------ 01:145038 1FC9 018 07-0008-06368E FC-3B00-06368E                07-1FC9-06368E  # noqa: E501
    # 047  I --- 01:145038 --:------ 01:145038 1FC9 018 FA-0008-06368E FC-3B00-06368E                FA-1FC9-06368E  # noqa: E501
    # 065  I --- 01:145038 --:------ 01:145038 1FC9 024 FC-0008-06368E FC-3150-06368E FB-3150-06368E FC-1FC9-06368E  # noqa: E501

    # HW valve binding:
    # 063  I --- 01:145038 --:------ 01:145038 1FC9 018 FA-0008-06368E FC-3B00-06368E FA-1FC9-06368E  # noqa: E501
    # CH valve binding:
    # 071  I --- 01:145038 --:------ 01:145038 1FC9 018 F9-0008-06368E FC-3B00-06368E F9-1FC9-06368E  # noqa: E501
    # ZoneValve zone binding
    # 045  W --- 13:106039 01:145038 --:------ 1FC9 012 00-3EF0-359E37 00-3B00-359E37
    # DHW binding..
    # 045  W --- 13:163733 01:145038 --:------ 1FC9 012 00-3EF0-367F95 00-3B00-367F95

    # 049  I --- 01:145038 --:------ 01:145038 1FC9 018 F9-0008-06368E FC-3B00-06368E F9-1FC9-06368E  # noqa

    # the new (heatpump-aware) BDR91:
    # 045 RP --- 13:035462 18:013393 --:------ 1FC9 018 00-3EF0-348A86 00-11F0-348A86 90-7FE1-DD6ABD # noqa

    def _parser(seqx) -> dict:
        # print(dev_hex_to_id(seqx[6:]))
        assert seqx[6:] == payload[6:12]  # all with same controller
        if seqx[:2] not in ("F9", "FA", "FB", "FC"):  # or: not in DOMAIN_TYPE_MAP: ??
            assert int(seqx[:2], 16) < MAX_ZONES
        return {seqx[:2]: seqx[2:6]}  # NOTE: codes is many:many (domain:code)

    assert msg.len >= 6 and msg.len % 6 == 0  # assuming not RQ
    assert msg.verb in (" I", " W", "RP")  # devices will respond to a RQ!
    assert msg.src.id == dev_hex_to_id(payload[6:12])
    return [
        _parser(payload[i : i + 12])
        for i in range(0, len(payload), 12)
        if payload[i : i + 2] != "90"  # WIP
    ]


@parser_decorator  # opentherm_sync
def parser_1fd4(payload, msg) -> Optional[dict]:
    assert msg.verb in " I"
    assert msg.len == 3
    assert payload[:2] == "00"

    return {"ticker": int(payload[2:], 16)}


@parser_decorator  # now_next_setpoint (non-Evohome, e.g. Sundial programmer)
def parser_2249(payload, msg) -> Optional[dict]:
    # see: https://github.com/jrosser/honeymon/blob/master/decoder.cpp#L357-L370
    # 095  I --- 23:100224 --:------ 23:100224 2249 007 00-7EFF-7EFF-FFFF
    # 095  I --- 23:100224 --:------ 23:100224 2249 007 00-7EFF-7EFF-FFFF

    def _parser(seqx) -> dict:
        return {
            **_idx(seqx[:2], msg),
            "setpoint_now": _temp(seqx[2:6]),
            "setpoint_next": _temp(seqx[6:10]),
            "unknown_0": int(seqx[10:], 16),  # countdown?
        }

    # the ST9520C can support two heating zones, so: msg.len in (7, 14)?
    if msg.is_array:  # TODO: can these msgs require >1 pkts? - seems unlikely
        assert msg.len >= 7 and msg.len % 7 == 0
        return [_parser(payload[i : i + 14]) for i in range(0, len(payload), 14)]

    assert msg.len == 7
    return _parser(payload)


@parser_decorator  # ufh_setpoint, TODO: max length = 24?
def parser_22c9(payload, msg) -> Optional[dict]:
    def _parser(seqx) -> dict:
        assert seqx[10:] == "01"

        return {
            **_idx(seqx[:2], msg),
            "temp_low": _temp(seqx[2:6]),
            "temp_high": _temp(seqx[6:10]),
            "_unknown_0": seqx[10:],
        }

    assert msg.len >= 6 and msg.len % 6 == 0
    return [_parser(payload[i : i + 12]) for i in range(0, len(payload), 12)]


@parser_decorator  # message_22d0 - system switch?
def parser_22d0(payload, msg) -> Optional[dict]:
    assert payload[:2] == "00"  # has no domain?
    assert payload[2:] == "000002"

    return {"unknown": payload[2:]}


@parser_decorator  # boiler_setpoint
def parser_22d9(payload, msg) -> Optional[dict]:
    assert msg.len == 3
    assert payload[:2] == "00"

    return {"boiler_setpoint": _temp(payload[2:6])}


@parser_decorator  # ???? (Nuaire 2 x 2-way switch)
def parser_22f1(payload, msg) -> Optional[dict]:
    # 11:42:43.149 081  I 051 --:------ --:------ 49:086353 22F1 003 000304
    # 11:42:49.587 071  I 052 --:------ --:------ 49:086353 22F1 003 000404
    # 11:42:49.685 072  I 052 --:------ --:------ 49:086353 22F1 003 000404
    # 11:42:49.784 072  I 052 --:------ --:------ 49:086353 22F1 003 000404
    assert msg.len == 3
    assert payload[:2] == "00"  # has no domain
    assert payload[4:] in ("04", "0A")

    bitmap = int(payload[2:4], 16)

    _bitmap = {"_bitmap": bitmap}

    if bitmap in (2, 3):
        _action = {"fan_mode": "normal" if bitmap == 2 else "boost"}
    elif bitmap in (9, 10):
        _action = {"heater_mode": "auto" if bitmap == 10 else "off"}
    else:
        _action = {}

    return {**_action, **_bitmap, "unknown_0": payload[4:]}


@parser_decorator  # similar to 22F1? switch?
def parser_22f3(payload, msg) -> Optional[dict]:
    assert msg.len == 3
    assert payload[:2] == "00"  # has no domain
    assert payload[4:6] == "0A"

    return {"_bitmap": int(payload[2:4], 16)}


@parser_decorator  # setpoint (of device/zones)
def parser_2309(payload, msg) -> Union[dict, list, None]:
    def _parser(seqx) -> dict:
        return {**_idx(seqx[:2], msg), "setpoint": _temp(seqx[2:])}

    # 055 RQ --- 12:010740 13:163733 --:------ 2309 003 0007D0
    # 046 RQ --- 12:010740 01:145038 --:------ 2309 003 03073A

    assert msg.verb in (" I", "RP", " W")

    if msg.is_array:
        assert msg.len >= 3 and msg.len % 3 == 0  # assuming not RQ
        return [_parser(payload[i : i + 6]) for i in range(0, len(payload), 6)]

    assert msg.len == 3
    return _parser(payload)


@parser_decorator  # zone_mode
def parser_2349(payload, msg) -> Optional[dict]:
    assert msg.verb in (" I", "RP", " W")
    assert msg.len in (7, 13)  # has a dtm if mode == "04"
    assert payload[6:8] in list(ZONE_MODE_MAP)
    assert payload[8:14] == "FFFFFF"

    result = {"mode": ZONE_MODE_MAP.get(payload[6:8]), "setpoint": _temp(payload[2:6])}
    if msg.len == 13:
        result.update({"until": _dtm(payload[14:26])})

    return {**_idx(payload[:2], msg), **result}


@parser_decorator  # hometronics _state (of unknwon)
def parser_2d49(payload, msg) -> dict:
    assert payload[:2] in ("88", "FD") or int(payload[:2], 16) < MAX_ZONES
    assert payload[2:] in ("0000", "C800")  # would "FFFF" mean N/A?
    # assert msg.len == 3  # implied

    return {**_idx(payload[:2], msg), "_state": _bool(payload[2:4])}


@parser_decorator  # system_mode
def parser_2e04(payload, msg) -> Optional[dict]:
    # if msg.verb == " W":
    # RQ/2E04/FF

    assert msg.len == 8
    assert payload[:2] in list(SYSTEM_MODE_MAP)  # TODO: check AutoWithReset

    return {
        "system_mode": SYSTEM_MODE_MAP.get(payload[:2]),
        "until": _dtm(payload[2:14]) if payload[14:] != "00" else None,
    }  # TODO: double-check the final "00"


@parser_decorator  # temperature (of device, zone/s)
def parser_30c9(payload, msg) -> Optional[dict]:
    def _parser(seqx) -> dict:
        return {**_idx(seqx[:2], msg), "temperature": _temp(seqx[2:])}

    if msg.is_array:
        assert msg.len >= 3 and msg.len % 3 == 0  # assuming not RQ
        return [_parser(payload[i : i + 6]) for i in range(0, len(payload), 6)]

    assert msg.len == 3
    return _parser(payload)


@parser_decorator  # unknown, from STA
def parser_3120(payload, msg) -> Optional[dict]:
    # sent by STAs every ~3:45:00, why?
    assert msg.src.type == "34"
    assert payload == "0070B0000000FF"
    return {"unknown_0": payload}


@parser_decorator  # datetime_sync
def parser_313f(payload, msg) -> Optional[dict]:
    # 2020-03-28T03:59:21.315178 045 RP --- 01:158182 04:136513 --:------ 313F 009 00FC3500A41C0307E4  # noqa: E501
    # 2020-03-29T04:58:30.486343 045 RP --- 01:158182 04:136485 --:------ 313F 009 00FC8400C51D0307E4  # noqa: E501
    # 2020-05-31T11:37:50.351511 056  I --- --:------ --:------ 12:207082 313F 009 0038021ECB1F0507E4  # noqa: E501

    # https://www.automatedhome.co.uk/vbulletin/showthread.php?5085-My-HGI80-equivalent-Domoticz-setup-without-HGI80&p=36422&viewfull=1#post36422
    # every day at ~4am TRV/RQ->CTL/RP, approx 5-10secs apart (CTL respond at any time)
    assert msg.len == 9
    assert payload[:2] == "00"  # evohome is always "00FC"?
    return {
        "datetime": _dtm(payload[4:18]),
        "is_dst": True if bool(int(payload[4:6], 16) & 0x80) else None,
        "_unknown_0": payload[2:4],
    }


@parser_decorator  # heat_demand (of device, FC domain)
def parser_3150(payload, msg) -> Optional[dict]:
    # event-driven, and periodically; FC domain is highest of all zones
    # TODO: all have a valid domain will UFC/CTL respond to an RQ, for FC, for a zone?

    def _parser(seqx) -> dict:
        # assert seqx[:2] == "FC" or (int(seqx[:2], 16) < MAX_ZONES)  # <5, 8 for UFC
        return {**_idx(seqx[:2], msg), "heat_demand": _percent(seqx[2:])}

    if msg.src.type == "02" and msg.is_array:  # TODO: hometronics only?
        return [_parser(payload[i : i + 4]) for i in range(0, len(payload), 4)]

    assert msg.len == 2  # msg.src.type in ("01","02","10","04")
    return _parser(payload)  # TODO: check UFC/FC is == CTL/FC


@parser_decorator  # ???
def parser_31d9(payload, msg) -> Optional[dict]:
    assert payload[2:4] in ("00", "06")
    assert payload[4:6] == "FF" or int(payload[4:6], 16) <= 200

    if msg.len == 3:  # usu: I -->20: (no seq#)
        return {
            **_idx(payload[:2], msg),
            "percent_1": _percent(payload[4:6]),
            "unknown_0": payload[2:4],
        }

    assert msg.len == 17  # usu: I 30:-->30:, (or 20:) with a seq#!
    assert payload[6:8] == "00"
    assert payload[8:32] in ("00" * 12, "20" * 12)

    return {
        **_idx(payload[:2], msg),
        "percent_1": _percent(payload[4:6]),
        "unknown_0": payload[2:4],
        "unknown_2": payload[6:8],
        "unknown_3": payload[8:32],
        "unknown_4": payload[32:],
    }


@parser_decorator  # UFC HCE80 (Nuaire humidity)
def parser_31da(payload, msg) -> Optional[dict]:
    assert msg.len == 29  # usu: I CTL-->CTL

    assert payload[2:10] == "EF007FFF"
    assert payload[12:30] == "EF7FFF7FFF7FFF7FFF"
    assert payload[34:36] == "EF"
    assert payload[42:44] == "00"
    assert payload[46:48] in ("00", "EF")
    assert payload[48:] in ("EF7FFF7FFF", "EF7FFFFFFF")

    rh = int(payload[10:12], 16) / 100 if payload[10:12] != "EF" else None  # not /200!

    return {
        **_idx(payload[:2], msg),
        "relative_humidity": rh,
        "unknown_1": payload[30:32],
        "unknown_2": payload[32:34],
        "unknown_3": payload[36:38],
        "unknown_4": payload[38:40],
        "unknown_5": payload[44:46],
    }


@parser_decorator  # ???? (Nuaire on/off)
def parser_31e0(payload, msg) -> Optional[dict]:
    # cat pkts.log | grep 31DA | grep -v ' I ' (event-driven ex 168090, humidity sensor)
    # 11:09:49.973 045  I --- VNT:168090 GWY:082155  --:------ 31E0 004 00 00 00 00
    # 11:14:46.168 045  I --- VNT:168090 GWY:082155  --:------ 31E0 004 00 00 C8 00
    # TODO: track humidity against 00/C8, OR HEATER?

    assert msg.len == 4  # usu: I VNT->GWY
    assert payload[:4] == "0000"  # domain?
    assert payload[4:] in ("0000", "C800")

    return {
        "state_31e0": _bool(payload[4:6]),
        "unknown_0": payload[:4],
        "unknown_1": payload[6:],
    }


@parser_decorator  # opentherm_msg
def parser_3220(payload, msg) -> Optional[dict]:
    assert msg.len == 5
    assert payload[:2] == "00"

    # these are OpenTherm-specific assertions
    assert int(payload[2:4], 16) // 0x80 == parity(int(payload[2:], 16) & 0x7FFFFFFF)

    ot_msg_type = int(payload[2:4], 16) & 0x70
    assert ot_msg_type in OPENTHERM_MSG_TYPE

    assert int(payload[2:4], 16) & 0x0F == 0

    ot_msg_id = int(payload[4:6], 16)
    assert str(ot_msg_id) in OPENTHERM_MESSAGES["messages"]

    message = OPENTHERM_MESSAGES["messages"].get(str(ot_msg_id))

    result = {
        "id": payload[4:6],  # ot_msg_id,
        "msg_name": message["en"],
        "msg_type": OPENTHERM_MSG_TYPE[ot_msg_type],
    }

    if not message:
        return {**result, "value_raw": payload[6:]}

    if msg.verb == "RQ":
        assert ot_msg_type < 48
        assert payload[6:10] == "0000"
        return {
            **result,
            # "description": message["en"]
        }

    assert ot_msg_type > 48

    if isinstance(message["var"], dict):
        if isinstance(message["val"], dict):
            result["value_hb"] = ot_msg_value(
                payload[6:8], message["val"].get("hb", message["val"])
            )
            result["value_lb"] = ot_msg_value(
                payload[8:10], message["val"].get("lb", message["val"])
            )
        else:
            result["value_hb"] = ot_msg_value(payload[6:8], message["val"])
            result["value_lb"] = ot_msg_value(payload[8:10], message["val"])

    else:
        if message["val"] in ("flag8", "u8", "s8"):
            result["value"] = ot_msg_value(payload[6:8], message["val"])
        else:
            result["value"] = ot_msg_value(payload[6:10], message["val"])

    return {
        **result,
        # "description": message["en"],
    }


@parser_decorator  # sync_tpi (TPI cycle HB/sync)
def parser_3b00(payload, msg) -> Optional[dict]:
    # https://www.domoticaforum.eu/viewtopic.php?f=7&t=5806&start=105#p73681
    # TODO: alter #cycles/hour & check interval between 3B00/3EF0 changes
    """Decode a 3B00 packet (sync_tpi).

    The heat relay regularly broadcasts a 3B00 at the start (or the end?) of every TPI
    cycle, the frequency of which is determined by the (TPI) cycle rate in 1100.

    The CTL subsequently broadcasts a 3B00 (i.e. at the start of every TPI cycle).

    The OTB does not send these packets, but the CTL sends a regular broadcast
    anyway.
    """

    # 053  I --- 13:209679 --:------ 13:209679 3B00 002 00C8
    # 045  I --- 01:158182 --:------ 01:158182 3B00 002 FCC8
    # 052  I --- 13:209679 --:------ 13:209679 3B00 002 00C8
    # 045  I --- 01:158182 --:------ 01:158182 3B00 002 FCC8

    # 063  I --- 01:078710 --:------ 01:078710 3B00 002 FCC8
    # 064  I --- 01:078710 --:------ 01:078710 3B00 002 FCC8

    assert msg.len == 2
    assert payload[:2] in {"01": "FC", "13": "00", "23": "FC"}.get(msg.src.type, "")
    assert payload[2:] == "C8"  # Could it be a percentage?

    return {**_idx(payload[:2], msg), "sync_tpi": _bool(payload[2:])}


@parser_decorator  # actuator_enabled (state)
def parser_3ef0(payload, msg) -> dict:
    # 045 RP --- 10:067219 01:078710 --:------ 3EF0 006 00 3C 10 0000FF
    # 074 RP --- 10:067219 01:078710 --:------ 3EF0 006 00 3C 10 0000FF

    # --- RP --- 10:138822 01:187666 --:------ 3EF0 006 00 00 10 0200FF
    # 063 RP --- 10:138822 01:187666 --:------ 3EF0 006 00 01 10 0200FF

    # 066 RP --- 10:067219 01:078710 --:------ 3EF0 006 00 00 10 0A00FF
    # 068 RP --- 10:067219 01:078710 --:------ 3EF0 006 00 2F 10 0A00FF
    # 066 RP --- 10:067219 01:078710 --:------ 3EF0 006 00 2F 10 0A00FF
    # 066 RP --- 10:067219 01:078710 --:------ 3EF0 006 00 2F 10 0A00FF
    # 069 RP --- 10:067219 01:078710 --:------ 3EF0 006 00 1D 10 0A00FF
    # 070 RP --- 10:067219 01:078710 --:------ 3EF0 006 00 12 10 0A00FF
    # 071 RP --- 10:067219 01:078710 --:------ 3EF0 006 00 11 10 0000FF

    # 072 RP --- 10:067219 01:078710 --:------ 3EF0 006 00 00 10 0000FF
    # 072 RP --- 10:067219 01:078710 --:------ 3EF0 006 00 00 10 0A00FF
    # 074 RP --- 10:067219 01:078710 --:------ 3EF0 006 00 00 10 0000FF

    # 095 RP --- 10:139656 34:212252 --:------ 3EF0 006 00 00 11 0000FF
    # 095 RP --- 10:114131 34:254475 --:------ 3EF0 006 00 00 10 000000

    # 060 RP --- 10:138822 01:187666 --:------ 3EF0 006 00 64 10 0C00FF
    # 058 RP --- 10:138822 01:187666 --:------ 3EF0 006 00 00 10 0400FF
    # 061 RP --- 10:138822 01:187666 --:------ 3EF0 006 00 64 10 0800FF

    # 063 RP --- 10:138822 01:187666 --:------ 3EF0 006 00 01 11 0100FF
    # 058 RP --- 10:138822 01:187666 --:------ 3EF0 006 00 00 11 0100FF
    # 057 RP --- 10:138822 01:187666 --:------ 3EF0 006 00 01 10 FA00FF
    # 062 RP --- 10:138822 01:187666 --:------ 3EF0 006 00 00 11 0100FF
    # 065 RP --- 10:138822 01:187666 --:------ 3EF0 006 00 FF 10 0200FF
    # 060 RP --- 10:138822 01:187666 --:------ 3EF0 006 00 00 11 0100FF
    # 062 RP --- 10:067219 01:078710 --:------ 3EF0 006 00 11 10 0AFFFF

    assert payload[:2] == "00"
    assert payload[-2:] == "FF"

    if msg.src.type == "10":  # OTB, to 01:, or 34:
        assert msg.len == 6
        assert int(payload[2:4], 16) <= 100 or payload[2:4] == "FF"
        assert payload[4:6] in ("10", "11")
        assert payload[6:8] in ("00", "01", "02", "04", "08", "0A", "0C")
        assert payload[8:12] == "00FF"  # or: in ("0000", "00FF", "FFFF")

        # TODO: this is WIP
        # if payload[6:8] == "0A":
        #     assert payload[2:4] != "00"
        # else:
        #     assert payload[2:4] == "00"

        # TODO: max value of [6:8] seen is "64", should be /200?
        return {
            **_idx(payload[:2], msg),
            "flame_state": payload[6:8],
            "unknown_0": payload[4:6],
            "flame_active": {"0A": True}.get(payload[6:8], False),
            "modulation_level": int(payload[2:4], 16) / 100,
        }

    # 051  I --- 13:049225 --:------ 13:049225 3EF0 003 00 00 FF
    # 054  I --- 13:209679 --:------ 13:209679 3EF0 003 00 C8 FF

    assert msg.len == 3
    assert payload[2:4] in ("00", "C8")
    assert payload[4:] == "FF"

    return {
        **_idx(payload[:2], msg),
        "actuator_enabled": _bool(payload[2:4]),  # TODO: is bool() too restrictive?
    }


@parser_decorator  # actuator_state
def parser_3ef1(payload, msg) -> dict:
    assert msg.verb == "RP"
    assert msg.len == 7
    assert payload[:2] == "00"
    assert payload[10:] in ("00FF", "C8FF")
    assert payload[-2:] == "FF"

    return {
        **_idx(payload[:2], msg),
        "actuator_enabled": _bool(payload[10:12]),
        "actuator_countdown": int(payload[6:10], 16),
        "cycle_countdown": int(payload[2:6], 16),
    }


@parser_decorator
def parser_unknown(payload, msg) -> Optional[dict]:
    # TODO: it may be useful to search payloads for hex_ids, commands, etc.
    raise NotImplementedError



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
"""The evohome-compatible devices."""

from datetime import datetime as dt, timedelta
import json
import logging
from typing import Any, Optional

from .command import Command, Priority
from .const import (
    __dev_mode__,
    # CODE_SCHEMA,
    # CODE_0005_ZONE_TYPE,
    # CODE_000C_DEVICE_TYPE,
    DEVICE_HAS_BATTERY,
    DEVICE_LOOKUP,
    DEVICE_TABLE,
    DEVICE_TYPES,
)
from .exceptions import CorruptStateError
from .logger import dt_now

_LOGGER = logging.getLogger(__name__)
if False and __dev_mode__:
    _LOGGER.setLevel(logging.DEBUG)
else:
    _LOGGER.setLevel(logging.WARNING)


def dev_hex_to_id(device_hex: str, friendly_id=False) -> str:
    """Convert (say) '06368E' to '01:145038' (or 'CTL:145038')."""
    if device_hex == "FFFFFE":  # aka '63:262142'
        return ">null dev<" if friendly_id else "63:262142"
    if not device_hex.strip():  # aka '--:------'
        return f"{'':10}" if friendly_id else "--:------"
    _tmp = int(device_hex, 16)
    dev_type = f"{(_tmp & 0xFC0000) >> 18:02d}"
    if friendly_id:
        dev_type = DEVICE_TYPES.get(dev_type, f"{dev_type:<3}")
    return f"{dev_type}:{_tmp & 0x03FFFF:06d}"


def dev_id_to_hex(device_id: str) -> str:
    """Convert (say) '01:145038' (or 'CTL:145038') to '06368E'."""
    if len(device_id) == 9:  # e.g. '01:123456'
        dev_type = device_id[:2]
    else:  # len(device_id) == 10, e.g. 'CTL:123456', or ' 63:262142'
        dev_type = DEVICE_LOOKUP.get(device_id[:3], device_id[1:3])
    return f"{(int(dev_type) << 18) + int(device_id[-6:]):0>6X}"  # sans preceding 0x


def _dtm(value) -> str:
    """Convert a datetime to a hex string."""

    def dtm_to_hex(tm_year, tm_mon, tm_mday, tm_hour, tm_min, tm_sec, *args):
        return f"{tm_min:02X}{tm_hour:02X}{tm_mday:02X}{tm_mon:02X}{tm_year:04X}"

    if value is None:
        return "FF" * 6

    if isinstance(value, str):
        try:
            value = dt.fromisoformat(value)
        except ValueError:
            raise ValueError("Invalid datetime isoformat string")
    elif not isinstance(value, dt):
        raise TypeError("Invalid datetime object")

    if value < dt_now() + timedelta(minutes=1):
        raise ValueError("Invalid datetime")

    return dtm_to_hex(*value.timetuple())


class Entity:
    """The Device/Zone base class."""

    def __init__(self, gateway, controller=None) -> None:
        self._gwy = gateway
        self._que = gateway.cmd_que
        self._ctl = controller

        self.id = None

        self._msgs = {}
        self._known_msg = None

    @property
    def controller(self):  # -> Optional[Controller]:
        """Return the id of the entity's controller, if known."""

        return self._ctl  # TODO: if the controller is not known, try to find it?

    @controller.setter
    def controller(self, controller) -> None:
        """Set the device's parent controller, after validating it."""

        if not isinstance(controller, Controller) and not controller.is_controller:
            raise TypeError(f"Not a controller: {controller}")

        if self._ctl is not None:  # zones have this set at instantiation
            if self._ctl is not controller:
                # 064  I --- 01:078710 --:------ 01:144246 1F09 003 FF04B5
                raise CorruptStateError(
                    f"Device {self} has a mismatched controller: "
                    f"old={self._ctl.id}, new={controller.id}",
                )
            return

        self._ctl = controller
        self._ctl.devices.append(self)
        self._ctl.device_by_id[self.id] = self
        _LOGGER.debug("Device %s: controller now set to %s", self.id, self._ctl.id)

    def _command(self, code, **kwargs) -> None:
        dest = kwargs.pop("dest_addr", self.id)
        verb = kwargs.pop("verb", "RQ")
        payload = kwargs.pop("payload", "00")

        self._msgs.pop(code, None)  # remove the old one, so we can tell if RP'd

        priority = Priority.HIGH if verb == " W" else Priority.DEFAULT
        kwargs["priority"] = kwargs.get("priority", priority)

        self._que.put_nowait(Command(verb, dest, code, payload, **kwargs))

    def _discover(self) -> None:
        raise NotImplementedError

    def _get_msg_value(self, code, key=None) -> dict:
        if self._msgs.get(code):
            if isinstance(self._msgs[code].payload, list):
                return self._msgs[code].payload

            if key is not None:
                return self._msgs[code].payload.get(key)

            result = self._msgs[code].payload
            return {
                k: v
                for k, v in result.items()
                if k[:1] != "_" and k not in ("domain_id", "zone_idx")
            }

    def _update_msg(self, msg) -> None:
        self._known_msg = msg.code in ("0016", "1FC9")  # TODO: push to zone?

        if "domain_id" in msg.payload:  # isinstance(msg.payload, dict) and
            self._domain[msg.payload["domain_id"]] = {msg.code: msg}  # 01/02/23
            return

        if msg.verb == " W":
            if msg.code in self._msgs and self._msgs[msg.code].verb != msg.verb:
                return

        if msg.verb == "RQ":  # and msg.payload:
            if msg.code in self._msgs and self._msgs[msg.code].verb != msg.verb:
                return

        # may get an RQ/W initially, but RP/I will override
        # self._msgs.update({msg.code: msg})
        self._msgs[msg.code] = msg

    @property
    def _pkt_codes(self) -> list:
        return list(self._msgs.keys())


class Actuator:  # 3EF0, 3EF1
    def __init__(self, gateway, device_addr, **kwargs) -> None:
        super().__init__(gateway, device_addr, **kwargs)

        self._actuator_enabled = None
        self._actuator_state = None

    def _discover(self) -> None:
        if self._gwy.config["disable_discovery"]:
            return

        super()._discover()

        self._command("3EF1")

    def _update_msg(self, msg) -> None:
        super()._update_msg(msg)

        if msg.code == "3EF0" and msg.verb == " I":
            self._actuator_enabled = msg.payload["actuator_enabled"]
            self._known_msg = True

        elif msg.code == "3EF1" and msg.verb == "RP":
            self._actuator_enabled = msg.payload["actuator_enabled"]
            self._actuator_state = msg.payload
            self._known_msg = True

    @property
    def actuator_enabled(self) -> Optional[bool]:  # 3EF0, TODO: does 10: RP/3EF1?
        return self._actuator_enabled

    @property
    def actuator_state(self) -> Optional[float]:  # 3EF1, TODO: not all actuators
        return self._actuator_state

    @property
    def status(self) -> dict:
        return {
            **super().status,
            "actuator_enabled": self.actuator_enabled,
            "actuator_state": self.actuator_state,
        }


class BatteryState:  # 1060
    def __init__(self, gateway, device_addr, **kwargs) -> None:
        super().__init__(gateway, device_addr, **kwargs)

        self._battery_state = {}

    def _update_msg(self, msg) -> None:
        super()._update_msg(msg)

        if msg.code == "1060" and msg.verb == " I":
            self._battery_state = msg.payload
            self._known_msg = True

    @property
    def battery_state(self) -> dict:  # 1060
        return self._battery_state

    @property
    def status(self) -> dict:
        return {**super().status, "battery_state": self.battery_state}


class HeatDemand:  # 3150
    def __init__(self, gateway, device_addr, **kwargs) -> None:
        super().__init__(gateway, device_addr, **kwargs)

        self._heat_demand = None

    def _update_msg(self, msg) -> None:
        super()._update_msg(msg)

        if msg.code == "3150" and msg.verb == " I":
            self._heat_demand = msg.payload["heat_demand"]
            self._known_msg = True

    @property
    def heat_demand(self) -> Optional[float]:  # 3150
        return self._heat_demand

    @property
    def status(self) -> dict:
        return {**super().status, "heat_demand": self.heat_demand}


class Setpoint:  # 2309
    def __init__(self, gateway, device_addr, **kwargs) -> None:
        super().__init__(gateway, device_addr, **kwargs)

        self._setpoint = None

    def _update_msg(self, msg) -> None:
        super()._update_msg(msg)

        if msg.code == "2309" and msg.verb in (" I", " W"):
            self._setpoint = msg.payload["setpoint"]
            self._known_msg = True

        elif msg.code == "2309" and msg.verb == "RQ":
            self._known_msg = True

    @property
    def setpoint(self) -> Optional[float]:  # 2309
        return self._setpoint

    @property
    def status(self) -> dict:
        return {**super().status, "setpoint": self.setpoint}


class Temperature:  # 30C9
    def __init__(self, gateway, device_addr, **kwargs) -> None:
        super().__init__(gateway, device_addr, **kwargs)

        self._temperature = None

    def _update_msg(self, msg) -> None:
        super()._update_msg(msg)

        if msg.code == "30C9" and msg.verb == " I":
            self._temperature = msg.payload["temperature"]
            self._known_msg = True

    @property
    def temperature(self) -> Optional[float]:  # 30C9
        return self._temperature

    @property
    def status(self) -> dict:
        return {**super().status, "temperature": self.temperature}


# ######################################################################################

# 00: used for unknown device types
class Device(Entity):
    """The Device base class."""

    def __init__(self, gateway, device_addr, controller=None, domain_id=None) -> None:
        _LOGGER.debug("Creating a Device: %s %s", device_addr.id, self.__class__)
        super().__init__(gateway, controller=controller)
        assert device_addr.id not in gateway.device_by_id, "Duplicate device address"

        self.id = device_addr.id
        self.hex_id = dev_id_to_hex(device_addr.id)

        gateway.devices.append(self)
        gateway.device_by_id[device_addr.id] = self

        if controller is not None:  # here, assumed to be valid
            controller.devices.append(self)
            controller.device_by_id[self.id] = self

        self.addr = device_addr
        self.type = device_addr.type

        if self.addr.type in DEVICE_TABLE:
            self._has_battery = DEVICE_TABLE[self.addr.type].get("has_battery")
            self._is_actuator = DEVICE_TABLE[self.addr.type].get("is_actuator")
            self._is_sensor = DEVICE_TABLE[self.addr.type].get("is_sensor")
        else:
            self._has_battery = None
            self._is_actuator = None
            self._is_sensor = None

        self._zone = None
        self._domain = {}
        self._domain_id = domain_id

        attrs = gateway.known_devices.get(device_addr.id)
        self._friendly_name = attrs.get("friendly_name") if attrs else None
        self._ignored = attrs.get("ignored", False) if attrs else False

        self._discover()

    def __repr__(self) -> str:
        """Return a complete representation of the device as a dict."""

        return f"{self.id} ({DEVICE_TYPES.get(self.type)})"

        result = {
            a: getattr(self, a)
            for a in dir(self)
            if not a.startswith("_") and not callable(getattr(self, a))
        }
        return json.dumps({self.id: result}, indent=2)

    def __str__(self) -> str:
        """Return a brief representation of the device as a string."""

        return f"{self.id} ({DEVICE_TYPES.get(self.type)})"

    def _discover(self) -> None:
        if self._gwy.config["disable_discovery"]:
            return

        # if self.id in ("01:145038", "13:035462"):
        # if self.type == "02":
        # if self.id == "02:001075":
        #     for code in CODE_SCHEMA:
        #         for payload in ("00", "0000", "01", "0100", "FF", "FC", "FB"):
        #             self._command(code, payload=payload)
        #     return

        # these won't respond to RQs unless in test/bind mode
        if self.type in DEVICE_HAS_BATTERY:
            return

        # do these even if battery-powered (e.g. device might be in rf_check mode)
        # if not __dev_mode__:
        for code in ("0016", "1FC9"):
            self._command(code, payload="0000" if code == "0016" else "00")

        # if not __dev_mode__ and self.has_battery is not True:
        self._command("10E0")

        # if self.addr.type not in ("01", "13") and not self.has_battery:  # TODO: dev
        #     for code in CODE_SCHEMA:
        #         if code == "0404":
        #             continue
        #         self._command(code, payload="0000" if code != "1F09" else "00")

    def _update_msg(self, msg) -> None:
        super()._update_msg(msg)

        self._known_msg = True
        if msg.code == "0016":
            pass  # self._rf_signal =
        elif msg.code == "10E0":
            pass  # self._hardware_info =
        elif msg.code == "1FC9":
            pass  # self._rf_level =
        else:
            self._known_msg = False

        # TODO: status updates always, but...
        # TODO: schema updates only if eavesdropping is enabled.
        if self._ctl is not None and "parent_idx" in msg.payload:
            self.zone = self._ctl.get_zone(msg.payload["parent_idx"])

    def _set_domain(self, ctl=None, dhw=None, zone=None) -> None:
        """Set the device's parent controller, after validating it."""

        if ctl is not None:
            self._domain_id = "FC"  # boiler_control

        elif dhw is not None:
            self._domain_id = "FA"
            ctl = dhw._ctl

        elif zone is not None:
            self._domain_id = zone.idx
            ctl = zone._ctl

        if self._ctl is None:  # zones have this set at instantiation
            self._ctl = ctl
            self._ctl.devices.append(self)
            self._ctl.device_by_id[self.id] = self
            _LOGGER.debug("Device %s: Controller now set to %s", self.id, self._ctl.id)

        elif self._ctl is not ctl:
            # 064  I --- 01:078710 --:------ 01:144246 1F09 003 FF04B5
            raise CorruptStateError(
                f"Device {self} has a mismatched Controller: "
                f"old={self._ctl.id}, new={ctl.id}",
            )

        if dhw is not None:
            self._zone = dhw
            _LOGGER.debug("Device %s: DhwZone now set to %s", self.id, dhw.id)
            return

        elif zone is None:
            return

        if self._zone is None:
            self._zone = zone
            self._zone.devices.append(self)
            self._zone.device_by_id[self.id] = self
            _LOGGER.debug("Device %s: Zone now set to %s", self.id, zone.id)

        elif self._zone is not zone:
            raise CorruptStateError(
                f"Device {self} has a mismatched Zone: "
                f"old={self._zone.idx}, new={zone.idx}",
            )

    @property
    def zone(self) -> Optional[Entity]:  # should be: Optional[Zone]
        """Return the device's parent zone, if known."""

        return self._zone

    @zone.setter
    def zone(self, zone: Entity) -> None:  # should be: zone: Zone
        """Set the device's parent zone, after validating it.

        There are three possible sources for the parent zone of a device:
        1. a 000C packet (from their controller) for actuators only
        2. a message.payload["zone_idx"]
        3. the sensor-matching algorithm fro zone sensors only

        All three will execute a dev.zone = zone (i.e. via this setter).

        Devices don't have parents, rather: Zones have children; a mis-configured
        system could have a device as a child of two domains.
        """

        if not isinstance(zone, Entity):  # should be: zone, Zone)
            raise TypeError(f"Not a zone: {zone}")

        if self._zone is not None:
            if self._zone is not zone:
                raise CorruptStateError(
                    f"Device {self} has a mismatched parent zone: "
                    f"old={self._zone}, new={zone}",
                )
            return

        self._domain_id = zone.idx
        self._zone = zone
        if self._domain_id == "FA":
            # if isinstance(self, DhwSensor):
            #     self._sensor = self
            # else:
            #     self._dhw_valve = self
            pass

        elif self not in self._zone.devices:
            self._zone.devices.append(self)
            self._zone.device_by_id[self.id] = self
        _LOGGER.debug("Device %s: parent zone now set to %s", self.id, self._zone)

    @property
    def description(self) -> Optional[str]:
        return DEVICE_TABLE[self.type]["name"] if self.type in DEVICE_TABLE else None

    @property
    def hardware_info(self) -> Optional[str]:  # 10E0
        return self._get_msg_value("10E0")

    @property
    def has_battery(self) -> Optional[bool]:  # 1060
        """Return True if a device is battery powered.

        Devices with a battery-backup may still be mains-powered.
        """
        if self._has_battery is not None:
            return self._has_battery

        if "1060" in self._msgs:
            self._has_battery = True
        return self._has_battery

    @property
    def is_controller(self) -> Optional[bool]:  # 1F09
        if self._ctl is self:
            return True
        # if isinstance(self, Controller):
        #     return True
        # if self.type in ("01", "23"):
        #     return True
        # if "1F09" in self._msgs:  # TODO: needs to add msg to instaition
        #     return self._msgs["1F09"].verb == " I"
        # if "31D9" in self._msgs:  # TODO: needs to add msg to instaition
        #     return self._msgs["31D9"].verb == " I"
        return False

    @property
    def _is_present(self) -> bool:
        return any([m.src.id == self.id for m in self._msgs.values()])

    @property
    def params(self):
        return {}

    @property
    def status(self):
        return {}


# 01:
class Controller(Device):
    """The Controller base class, supports child devices and zones only."""

    def __init__(self, gateway, device_addr, **kwargs) -> None:
        super().__init__(gateway, device_addr, **kwargs)

        self.devices = [self]
        self.device_by_id = {self.id: self}

        self._ctl = self
        # self._domain_id = "FF"


# 02: "10E0", "3150";; "0008", "22C9", "22D0"
class UfhController(Device):
    """The UFC class, the HCE80 that controls the UFH zones."""

    # 12:27:24.398 067  I --- 02:000921 --:------ 01:191718 3150 002 0360
    # 12:27:24.546 068  I --- 02:000921 --:------ 01:191718 3150 002 065A
    # 12:27:24.693 067  I --- 02:000921 --:------ 01:191718 3150 002 045C
    # 12:27:24.824 059  I --- 01:191718 --:------ 01:191718 3150 002 FC5C
    # 12:27:24.857 067  I --- 02:000921 --:------ 02:000921 3150 006 0060015A025C

    def __init__(self, gateway, device_addr, **kwargs) -> None:
        super().__init__(gateway, device_addr, **kwargs)

        self._circuits = {}

    def _discover(self) -> None:
        if self._gwy.config["disable_discovery"]:
            return

        super()._discover()

        # [  # 3150: no answer
        #     self._command("3150", payload=f"{zone_idx:02X}")for zone_idx in range(8)
        # ]

        # [  # 22C9: no answer
        #     self._command("22C9", payload=f"{payload}")
        #     for payload in ("00", "0000", "01", "0100")
        # ]

        # [  # 22D0: dunno, always: {'unknown': '000002'}
        #     self._command("22D0", payload=f"{payload}")
        #     for payload in ("00", "0000", "00000002")
        # ]

        [  # 000C: used to find evo zone for each configured channel
            self._command("000C", payload=f"{idx:02X}{dev_type}")
            for dev_type in ("09",)  # CODE_000C_DEVICE_TYPE, also ("00", "04")
            # for dev_type in CODE_000C_DEVICE_TYPE
            for idx in range(8)  # for each possible UFH channel
        ]

        # [  # 0005: shows which channels are active - ?no use? (see above)
        #     self._command("0005", payload=f"00{zone_type}")
        #     # for zone_type in ("09",)  # CODE_0005_ZONE_TYPE, also ("00", "04", "0F")
        #     for zone_type in CODE_0005_ZONE_TYPE
        # ]

    def _update_msg(self, msg) -> None:
        def do_3150_magic() -> None:
            return

        super()._update_msg(msg)

        #
        if self._known_msg:
            pass
        elif msg.code in (
            "0001",
            "0005",
            "0008",
            "000A",
            "000C",
            "22C9",
            "22D0",
            "2309",
            "3150",
        ):
            pass
        else:
            assert False, f"Unknown packet verb/code for {self.id}"

        # "0008|FA/FC", "22C9|array", "22D0|none", "3150|ZZ/array(/FC?)"

        if msg.code == "000C":
            if "ufh_idx" in msg.payload and msg.payload["zone_id"] is not None:
                self._circuits[msg.payload["ufh_idx"]] = {
                    "zone_idx": msg.payload["zone_id"]
                }

        # if msg.code in ("22C9") and not isinstance(msg.payload, list):
        #     pass
        # else:
        #     assert False, "Unknown packet code"

    @property
    def setpoints(self):  # 22C9
        return self._get_msg_value("22C9")

    @property  # id, type
    def schema(self) -> dict:
        schema = {"ufh_circuits": self._circuits}

        return schema

    # @property  # setpoint, config, mode (not schedule)
    # def params(self) -> dict:
    #     ATTR_NAME = "name"
    #     ATTR_MODE = "mode"
    #     ATTR_CONFIG = "zone_config"

    #     return {
    #         ATTR_NAME: self.name,
    #         ATTR_MODE: self.mode,
    #         ATTR_CONFIG: self.zone_config,
    #     }

    # @property
    # def status(self) -> dict:
    #     return {
    #         ATTR_SETPOINT: self.setpoint,
    #         ATTR_TEMP: self.temperature,
    #     }


# 07: "1060";; "1260" "10A0"
class DhwSensor(BatteryState, Device):
    """The DHW class, such as a CS92."""

    def __init__(self, gateway, device_addr, **kwargs) -> None:
        super().__init__(gateway, device_addr, **kwargs)

        self._domain_id = "FA"

        # self._dhw_params = {}
        self._temperature = None

    def _update_msg(self, msg) -> None:
        super()._update_msg(msg)

        if self._known_msg:
            pass
        elif msg.verb in ("RP", " W"):
            assert False, f"Unknown packet verb for {self.id}"
        elif msg.code in "1260" and msg.verb == " I":
            self._temperature = msg.payload["temperature"]
        elif msg.verb == "RQ" and msg.code == "10A0":
            self._dhw_params = msg.payload
        else:
            assert False, f"Unknown packet code for {self.id}"

    # @property
    # def dhw_params(self) -> dict:
    #     return self._dhw_params

    @property
    def temperature(self) -> Optional[float]:
        return self._temperature

    # @property
    # def params(self) -> dict:
    #     return {**super().params, "dhw_params": self.dhw_params}

    @property
    def status(self) -> dict:
        return {**super().status, "temperature": self.temperature}


# 10: "10E0", "3EF0", "3150";; "22D9", "3220" ("1FD4"), TODO: 3220
class OtbGateway(Actuator, HeatDemand, Device):
    """The OTB class, specifically an OpenTherm Bridge (R8810A Bridge)."""

    def __init__(self, gateway, device_addr, **kwargs) -> None:
        super().__init__(gateway, device_addr, **kwargs)

        self._domain_id = "FC"

    def _update_msg(self, msg) -> None:
        super()._update_msg(msg)

        if self._known_msg:
            pass
        else:
            assert False, f"Unknown packet code for {self.id}"

    @property
    def boiler_setpoint(self) -> Optional[Any]:  # 22D9
        return self._get_msg_value("22D9", "boiler_setpoint")

    @property
    def _last_opentherm_msg(self) -> Optional[Any]:  # 3220
        return self._get_msg_value("3220")

    @property
    def status(self) -> dict:
        return {**super().status, "boiler_setpoint": self.boiler_setpoint}


# 03/12/22/34: 1060/2309/30C9;; (03/22: 0008/0009/3EF1, 2349?) (34: 000A/10E0/3120)
class Thermostat(BatteryState, Setpoint, Temperature, Device):
    """The THM/STA class, such as a TR87RF."""

    def _update_msg(self, msg) -> None:
        super()._update_msg(msg)

        if self._known_msg:
            pass

        elif self.type == "03":
            if msg.verb == " I" and msg.code in ("0008", "0009", "1100"):
                pass
            else:
                assert False, f"Unknown packet code for {self.id}"

        elif self.type in ("12", "22"):
            if msg.verb == " I" and msg.code in (
                "0008",
                "0009",
                "1030",
                "1100",
                "313F",
            ):
                pass
            elif msg.verb == "RQ" and msg.code in ("000A", "3EF1"):
                pass
            elif msg.verb == " W" and msg.code in ("2349"):
                pass
            else:
                assert False, f"Unknown packet code for {self.id}"

        elif self.type == "34":
            if msg.verb == " I" and msg.code in ("042F", "10E0", "3120"):
                pass
            elif msg.verb == "RQ" and msg.code in ("000A",):
                pass
            else:
                assert False, f"Unknown packet code for {self.id}"


class BdrSwitch(Actuator, Device):
    """The BDR class, such as a BDR91.

    BDR91s can be used in six disctinct modes, including:
    - x2 boiler controller (FC/TPI): normal, and heat pump (new)
    - x1 electric heat zones (0x/ELE)
    - x1 zone valve zones (0x/VAL)
    - x2 DHW thingys (HW/xxx)
    """

    def __init__(self, gateway, device_addr, **kwargs) -> None:
        super().__init__(gateway, device_addr, **kwargs)

        self._relay_demand = None

        self._is_tpi = kwargs.get("domain_id") == "FC"
        if self._is_tpi:
            self._ctl.boiler_control = self

    def _discover(self) -> None:
        """The BDRs fail(?) to respond to RQs for: 3B00, 3EF0, 0009.

        They seem to respond to:
        - 0008: varies on/off
        - 1100
        - 3EF1: has sub-domains?
        """

        if self._gwy.config["disable_discovery"]:
            return

        super()._discover()

        # self._command("1100", payload="00")  # will just repeat the controller config
        self._command("0008")

    def _update_msg(self, msg) -> None:
        super()._update_msg(msg)

        if self._is_tpi is None:
            _ = self.is_tpi

        if self._known_msg:
            pass
        elif msg.code == "0008" and msg.verb == "RP":
            self._relay_demand = msg.payload["relay_demand"]
        elif msg.code == "1100":
            pass
        elif msg.code in "3B00":
            pass
        else:
            assert False, f"Unknown packet verb/code for {self.id}"

    @property
    def is_tpi(self) -> Optional[bool]:  # 3B00
        if self._is_tpi is not None:
            return self._is_tpi

        elif "1FC9" in self._msgs and self._msgs["1FC9"].verb == "RP":
            if "3B00" in self._msgs["1FC9"].raw_payload:
                self._is_tpi = True

        elif "3B00" in self._msgs and self._msgs["3B00"].verb == " I":
            self._is_tpi = True

        if self._is_tpi:
            self._domain_id = "FC"  # TODO: check is None first
            self.ctl.boiler_control = self

        return self._is_tpi

    @property
    def _tpi_params(self) -> dict:  # 1100
        return self._get_msg_value("1100")  # if self.is_tpi else None

    @property
    def relay_demand(self) -> dict:  # 0008
        return self._relay_demand


class TrvActuator(BatteryState, HeatDemand, Setpoint, Temperature, Device):
    """The TRV class, such as a HR92."""

    def __init__(self, gateway, device_addr, **kwargs) -> None:
        super().__init__(gateway, device_addr, **kwargs)
        self._window_state = None

    def _update_msg(self, msg) -> None:
        super()._update_msg(msg)

        if self._known_msg:
            pass
        elif msg.code in "12B0" and msg.verb == " I":
            self._window_state = msg.payload["window_open"]
        elif msg.verb == "RQ" and msg.code in ("0100", "313F"):
            pass
        else:
            assert False, f"Unknown packet verb/code for {self.id}"

    @property
    def window_state(self) -> Optional[bool]:  # 12B0
        return self._window_state

    @property
    def status(self) -> dict:
        return {**super().status, "window_state": self.window_state}


DEVICE_CLASSES = {
    "01": Controller,  # use EvoSystem instead?
    "02": UfhController,
    "03": Thermostat,
    "04": TrvActuator,
    "07": DhwSensor,
    "10": OtbGateway,
    "12": Thermostat,
    "13": BdrSwitch,
    "22": Thermostat,
    "23": Controller,  # a Programmer, use System instead?
    "34": Thermostat,
}
